#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Ansible Project
# Copyright (c) 2022 Alina Buzachis (@alinabuzachis)
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: rds_cluster
version_added: 5.0.0
short_description: rds_cluster module
description:
  - Create, modify, and delete RDS clusters.
  - This module was originally added to C(community.aws) in release 3.2.0.
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
author:
  - Sloane Hertel (@s-hertel)
  - Alina Buzachis (@alinabuzachis)
options:
  # General module options
    state:
        description:
          - Whether the snapshot should exist or not.
          - V(started) and V(stopped) can only be used with aurora clusters.
          - Support for V(started) and V(stopped) was added in release 6.3.0.
        choices: ['present', 'absent', 'started', 'stopped']
        default: 'present'
        type: str
    creation_source:
        description: Which source to use if creating from a template (an existing cluster, S3 bucket, or snapshot).
        choices: ['snapshot', 's3', 'cluster']
        type: str
    force_update_password:
        description:
          - Set to V(true) to update your cluster password with O(master_user_password).
          - Since comparing passwords to determine if it needs to be updated is not possible this is set to V(false) by default to allow idempotence.
        type: bool
        default: false
    promote:
        description: Set to V(true) to promote a read replica cluster.
        type: bool
        default: false
    purge_cloudwatch_logs_exports:
        description:
          - Whether or not to disable Cloudwatch logs enabled for the DB cluster that are not provided in O(enable_cloudwatch_logs_exports).
            Set O(enable_cloudwatch_logs_exports) to an empty list to disable all.
        type: bool
        default: true
    purge_security_groups:
        description:
          - Set to V(false) to retain any enabled security groups that aren't specified in the task and are associated with the cluster.
          - Can be applied to O(vpc_security_group_ids)
        type: bool
        default: true
    wait:
        description: Whether to wait for the cluster to be available or deleted.
        type: bool
        default: true
    # Options that have a corresponding boto3 parameter
    apply_immediately:
        description:
          - A value that specifies whether modifying a cluster with O(new_db_cluster_identifier) and O(master_user_password)
            should be applied as soon as possible, regardless of the O(preferred_maintenance_window) setting. If V(false), changes
            are applied during the next maintenance window.
        type: bool
        default: false
    availability_zones:
        description:
          - A list of EC2 Availability Zones that instances in the DB cluster can be created in.
            May be used when creating a cluster or when restoring from S3 or a snapshot.
        aliases:
          - zones
          - az
        type: list
        elements: str
    backtrack_to:
        description:
          - The timestamp of the time to backtrack the DB cluster to in ISO 8601 format, such as "2017-07-08T18:00Z".
        type: str
    backtrack_window:
        description:
          - The target backtrack window, in seconds. To disable backtracking, set this value to V(0).
          - If specified, this value must be set to a number from V(0) to V(259,200) (72 hours).
        type: int
    backup_retention_period:
        description:
          - The number of days for which automated backups are retained (must be within V(1) to V(35)).
            May be used when creating a new cluster, when restoring from S3, or when modifying a cluster.
        type: int
        default: 1
    character_set_name:
        description:
          - The character set to associate with the DB cluster.
        type: str
    database_name:
        description:
          - The name for your database. If a name is not provided Amazon RDS will not create a database.
        aliases:
          - db_name
        type: str
    db_cluster_identifier:
        description:
          - The DB cluster (lowercase) identifier. The identifier must contain from 1 to 63 letters, numbers, or
            hyphens and the first character must be a letter and may not end in a hyphen or contain consecutive hyphens.
        aliases:
          - cluster_id
          - id
          - cluster_name
        type: str
        required: true
    db_cluster_parameter_group_name:
        description:
          - The name of the DB cluster parameter group to associate with this DB cluster.
            If this argument is omitted when creating a cluster, the default DB cluster parameter group for the specified DB engine and version is used.
        type: str
    db_subnet_group_name:
        description:
          - A DB subnet group to associate with this DB cluster if not using the default.
        type: str
    enable_cloudwatch_logs_exports:
        description:
          - A list of log types that need to be enabled for exporting to CloudWatch Logs.
          - Engine aurora-mysql supports V(audit), V(error), V(general) and V(slowquery).
          - Engine aurora-postgresql supports V(postgresql).
        type: list
        elements: str
    deletion_protection:
        description:
          -  A value that indicates whether the DB cluster has deletion protection enabled.
             The database can't be deleted when deletion protection is enabled.
             By default, deletion protection is disabled.
        type: bool
    global_cluster_identifier:
        description:
          -  The global cluster ID of an Aurora cluster that becomes the primary cluster in the new global database cluster.
        type: str
    enable_http_endpoint:
        description:
          -  A value that indicates whether to enable the HTTP endpoint for an Aurora Serverless DB cluster.
             By default, the HTTP endpoint is disabled.
        type: bool
    copy_tags_to_snapshot:
        description:
          - Indicates whether to copy all tags from the DB cluster to snapshots of the DB cluster.
            The default is not to copy them.
        type: bool
    domain:
        description:
          - The Active Directory directory ID to create the DB cluster in.
        type: str
    domain_iam_role_name:
        description:
          - Specify the name of the IAM role to be used when making API calls to the Directory Service.
        type: str
    enable_global_write_forwarding:
        description:
          - A value that indicates whether to enable this DB cluster to forward write operations to the primary cluster of an Aurora global database.
            By default, write operations are not allowed on Aurora DB clusters that are secondary clusters in an Aurora global database.
          - This value can be only set on Aurora DB clusters that are members of an Aurora global database.
        type: bool
    db_cluster_instance_class:
        description:
          - The compute and memory capacity of each DB instance in the Multi-AZ DB cluster, for example V(db.m6gd.xlarge).
          - Not all DB instance classes are available in all Amazon Web Services Regions, or for all database engines.
          - For the full list of DB instance classes and availability for your engine visit
            U(https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.DBInstanceClass.html).
          - This setting is required to create a Multi-AZ DB cluster.
        type: str
        version_added: 5.5.0
    enable_iam_database_authentication:
        description:
          - Enable mapping of AWS Identity and Access Management (IAM) accounts to database accounts.
            If this option is omitted when creating the cluster, Amazon RDS sets this to C(false).
        type: bool
    allocated_storage:
        description:
          - The amount of storage in gibibytes (GiB) to allocate to each DB instance in the Multi-AZ DB cluster.
          - This setting is required to create a Multi-AZ DB cluster.
        type: int
        version_added: 5.5.0
    storage_type:
        description:
          - Specifies the storage type to be associated with the DB cluster.
          - This setting is required to create a Multi-AZ DB cluster.
          - For multi-AZ DB clusters, O(storage_type) defaults to V(io1) and a value for the O(iops) parameter is required.
          - For Aurora DB clusters, O(storage_type) defaults to V(aurora) standard.
          - For mysql and postgres DB clusters, O(storage_type) defaults to V(io1).
          - Support for V(aurora) and V(aurora-iopt1) was added in release 8.2.0.
        type: str
        choices:
          - io1
          - aurora
          - aurora-iopt1
        version_added: 5.5.0
    iops:
        description:
          - The amount of Provisioned IOPS (input/output operations per second) to be initially allocated for each DB instance in the Multi-AZ DB cluster.
          - This setting is required to create a Multi-AZ DB cluster
          - Must be a multiple between .5 and 50 of the storage amount for the DB cluster.
        type: int
        version_added: 5.5.0
    engine:
        description:
          - The name of the database engine to be used for this DB cluster. This is required to create a cluster.
          - The combinaison of I(engine) and I(engine_mode) may not be supported.
          - "See AWS documentation for details:
            L(Amazon RDS Documentation,https://docs.aws.amazon.com/AmazonRDS/latest/APIReference/API_CreateDBCluster.html)."
          - When O(engine=mysql), O(allocated_storage), O(iops) and O(db_cluster_instance_class) must also be specified.
          - When I(engine=postgres), O(allocated_storage), O(iops) and O(db_cluster_instance_class) must also be specified.
          - Support for V(postgres) and V(mysql) was added in amazon.aws 5.5.0.
        choices:
          - aurora
          - aurora-mysql
          - aurora-postgresql
          - mysql
          - postgres
        type: str
    engine_mode:
        description:
          - The DB engine mode of the DB cluster. The combination of O(engine) and O(engine_mode) may not be supported.
          - "See AWS documentation for details:
            L(Amazon RDS Documentation,https://docs.aws.amazon.com/AmazonRDS/latest/APIReference/API_CreateDBCluster.html)."
        choices:
          - provisioned
          - serverless
          - parallelquery
          - global
          - multimaster
        type: str
        version_added: 5.5.0
    engine_version:
        description:
          - The version number of the database engine to use.
          - For Aurora MySQL that could be V(5.6.10a), V(5.7.12).
          - Aurora PostgreSQL example, V(9.6.3).
        type: str
    final_snapshot_identifier:
        description:
          - The DB cluster snapshot identifier of the new DB cluster snapshot created when O(skip_final_snapshot=false).
        type: str
    force_backtrack:
        description:
          - A boolean to indicate if the DB cluster should be forced to backtrack when binary logging is enabled.
            Otherwise, an error occurs when binary logging is enabled.
        type: bool
    kms_key_id:
        description:
          - The AWS KMS key identifier (the ARN, unless you are creating a cluster in the same account that owns the
            KMS key, in which case the KMS key alias may be used).
          - If O(replication_source_identifier) specifies an encrypted source Amazon RDS will use the key used toe encrypt the source.
          - If O(storage_encrypted=true) and and O(replication_source_identifier) is not provided, the default encryption key is used.
        type: str
    master_user_password:
        description:
          - An 8-41 character password for the master database user.
          - The password can contain any printable ASCII character except V(/), V("), or V(@).
          - To modify the password use O(force_update_password). Use O(apply_immediately) to change
            the password immediately, otherwise it is updated during the next maintenance window.
        aliases:
          - password
        type: str
    master_username:
        description:
          - The name of the master user for the DB cluster. Must be 1-16 letters or numbers and begin with a letter.
        aliases:
          - username
        type: str
    new_db_cluster_identifier:
        description:
          - The new DB cluster (lowercase) identifier for the DB cluster when renaming a DB cluster.
          - The identifier must contain from 1 to 63 letters, numbers, or hyphens and the first character must be a
            letter and may not end in a hyphen or contain consecutive hyphens.
          - Use O(apply_immediately) to rename immediately, otherwise it is updated during the next maintenance window.
        aliases:
          - new_cluster_id
          - new_id
          - new_cluster_name
        type: str
    option_group_name:
        description:
          - The option group to associate with the DB cluster.
        type: str
    port:
        description:
          - The port number on which the instances in the DB cluster accept connections. If not specified, Amazon RDS
            defaults this to V(3306) if the O(engine) is V(aurora) and V(5432) if the O(engine) is V(aurora-postgresql).
        type: int
    preferred_backup_window:
        description:
          - The daily time range (in UTC) of at least 30 minutes, during which automated backups are created if automated backups are
            enabled using O(backup_retention_period). The option must be in the format of "hh24:mi-hh24:mi" and not conflict with
            O(preferred_maintenance_window).
        aliases:
          - backup_window
        type: str
    preferred_maintenance_window:
        description:
          - The weekly time range (in UTC) of at least 30 minutes, during which system maintenance can occur. The option must
            be in the format "ddd:hh24:mi-ddd:hh24:mi" where ddd is one of Mon, Tue, Wed, Thu, Fri, Sat, Sun.
        aliases:
          - maintenance_window
        type: str
    remove_from_global_db:
        description:
          - If set to V(true), the cluster will be removed from global DB.
          - Parameters O(global_cluster_identifier), O(db_cluster_identifier) must be specified when O(remove_from_global_db=true).
        type: bool
        required: false
        version_added: 6.5.0
    replication_source_identifier:
        description:
          - The Amazon Resource Name (ARN) of the source DB instance or DB cluster if this DB cluster is created as a Read Replica.
        aliases:
          - replication_src_id
        type: str
    restore_to_time:
        description:
          - The UTC date and time to restore the DB cluster to. Must be in the format "2015-03-07T23:45:00Z".
          - If this is not provided while restoring a cluster, O(use_latest_restorable_time) must be.
            May not be specified if O(restore_type) is copy-on-write.
        type: str
    restore_type:
        description:
          - The type of restore to be performed. If not provided, Amazon RDS uses full-copy.
        choices:
          - full-copy
          - copy-on-write
        type: str
    role_arn:
        description:
          - The Amazon Resource Name (ARN) of the IAM role to associate with the Aurora DB cluster, for example
            "arn:aws:iam::123456789012:role/AuroraAccessRole"
        type: str
    s3_bucket_name:
        description:
          - The name of the Amazon S3 bucket that contains the data used to create the Amazon Aurora DB cluster.
        type: str
    s3_ingestion_role_arn:
        description:
          - The Amazon Resource Name (ARN) of the AWS Identity and Access Management (IAM) role that authorizes Amazon RDS to access
            the Amazon S3 bucket on your behalf.
        type: str
    s3_prefix:
        description:
          - The prefix for all of the file names that contain the data used to create the Amazon Aurora DB cluster.
          - If you do not specify a SourceS3Prefix value, then the Amazon Aurora DB cluster is created by using all of the files in the Amazon S3 bucket.
        type: str
    serverless_v2_scaling_configuration:
        description:
          - Contains the scaling configuration of an Aurora Serverless v2 DB cluster.
        type: dict
        suboptions:
          min_capacity:
            description:
              - The minimum number of Aurora capacity units (ACUs) for a DB instance in an Aurora Serverless v2 cluster.
              - ACU values can be specified in in half-step increments, such as V(8), V(8.5), V(9), and so on.
              - The smallest possible value is V(0.5).
            type: float
          max_capacity:
            description:
              - The maximum number of Aurora capacity units (ACUs) for a DB instance in an Aurora Serverless v2 cluster.
              - ACU values can be specified in in half-step increments, such as V(40), V(40.5), V(41), and so on.
              - The largest possible value is V(128).
            type: float
        version_added: 7.3.0
    skip_final_snapshot:
        description:
          - Whether a final DB cluster snapshot is created before the DB cluster is deleted.
          - If this is V(false), O(final_snapshot_identifier) must be provided.
        type: bool
        default: false
    snapshot_identifier:
        description:
          - The identifier for the DB snapshot or DB cluster snapshot to restore from.
          - You can use either the name or the ARN to specify a DB cluster snapshot. However, you can use only the ARN to specify a DB snapshot.
        type: str
    source_db_cluster_identifier:
        description:
          - The identifier of the source DB cluster from which to restore.
        type: str
    source_engine:
        description:
          - The identifier for the database engine that was backed up to create the files stored in the Amazon S3 bucket.
        choices:
          - mysql
        type: str
    source_engine_version:
        description:
          - The version of the database that the backup files were created from.
        type: str
    source_region:
        description:
          - The ID of the region that contains the source for the DB cluster.
        type: str
    storage_encrypted:
        description:
          - Whether the DB cluster is encrypted.
        type: bool
    use_earliest_time_on_point_in_time_unavailable:
        description:
          - If O(backtrack_to) is set to a timestamp earlier than the earliest backtrack time, this value backtracks the DB cluster to
            the earliest possible backtrack time. Otherwise, an error occurs.
        type: bool
    use_latest_restorable_time:
        description:
          - Whether to restore the DB cluster to the latest restorable backup time. Only one of O(use_latest_restorable_time)
            and O(restore_to_time) may be provided.
        type: bool
    vpc_security_group_ids:
        description:
          - A list of EC2 VPC security groups to associate with the DB cluster.
        type: list
        elements: str
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.
- name: Create minimal Aurora cluster in default VPC and default subnet group
  amazon.aws.rds_cluster:
    cluster_id: "{{ cluster_id }}"
    engine: "aurora"
    password: "{{ password }}"
    username: "{{ username }}"

- name: Add a new security group without purge
  amazon.aws.rds_cluster:
    id: "{{ cluster_id }}"
    state: present
    vpc_security_group_ids:
      - sg-0be17ba10c9286b0b
    purge_security_groups: false

- name: Modify password
  amazon.aws.rds_cluster:
    id: "{{ cluster_id }}"
    state: present
    password: "{{ new_password }}"
    force_update_password: true
    apply_immediately: true

- name: Rename the cluster
  amazon.aws.rds_cluster:
    engine: aurora
    password: "{{ password }}"
    username: "{{ username }}"
    cluster_id: "cluster-{{ resource_prefix }}"
    new_cluster_id: "cluster-{{ resource_prefix }}-renamed"
    apply_immediately: true

- name: Delete Aurora cluster without creating a final snapshot
  amazon.aws.rds_cluster:
    engine: aurora
    password: "{{ password }}"
    username: "{{ username }}"
    cluster_id: "{{ cluster_id }}"
    skip_final_snapshot: true
    tags:
      Name: "cluster-{{ resource_prefix }}"
      Created_By: "Ansible_rds_cluster_integration_test"
    state: absent

- name: Restore cluster from source snapshot
  amazon.aws.rds_cluster:
    engine: aurora
    password: "{{ password }}"
    username: "{{ username }}"
    cluster_id: "cluster-{{ resource_prefix }}-restored"
    snapshot_identifier: "cluster-{{ resource_prefix }}-snapshot"

- name: Create an Aurora PostgreSQL cluster and attach an intance
  amazon.aws.rds_cluster:
    state: present
    engine: aurora-postgresql
    engine_mode: provisioned
    cluster_id: '{{ cluster_id }}'
    username: '{{ username }}'
    password: '{{ password }}'

- name: Attach a new instance to the cluster
  amazon.aws.rds_instance:
    id: '{{ instance_id }}'
    cluster_id: '{{ cluster_id }}'
    engine: aurora-postgresql
    state: present
    db_instance_class: 'db.t3.medium'

- name: Remove a cluster from global DB (do not delete)
  amazon.aws.rds_cluster:
    db_cluster_identifier: '{{ cluster_id }}'
    global_cluster_identifier: '{{ global_cluster_id }}'
    remove_from_global_db: true

- name: Remove a cluster from global DB and Delete without creating a final snapshot
  amazon.aws.rds_cluster:
    engine: aurora
    password: "{{ password }}"
    username: "{{ username }}"
    cluster_id: "{{ cluster_id }}"
    skip_final_snapshot: true
    remove_from_global_db: true
    wait: true
    state: absent

- name: Update cluster port and WAIT for remove secondary DB cluster from global DB to complete
  amazon.aws.rds_cluster:
    db_cluster_identifier: "{{ secondary_cluster_name }}"
    global_cluster_identifier: "{{ global_cluster_name }}"
    remove_from_global_db: true
    state: present
    port: 3389
    region: "{{ secondary_cluster_region }}"

- name: Update cluster port and DO NOT WAIT for remove secondary DB cluster from global DB to complete
  amazon.aws.rds_cluster:
    db_cluster_identifier: "{{ secondary_cluster_name }}"
    global_cluster_identifier: "{{ global_cluster_name }}"
    remove_from_global_db: true
    state: present
    port: 3389
    region: "{{ secondary_cluster_region }}"
    wait: false
"""

RETURN = r"""
activity_stream_status:
  description: The status of the database activity stream.
  returned: always
  type: str
  sample: "stopped"
allocated_storage:
  description:
    - The allocated storage size in gigabytes. Since aurora storage size is not fixed this is
      always 1 for aurora database engines.
  returned: always
  type: int
  sample: 1
associated_roles:
  description:
    - A list of dictionaries of the AWS Identity and Access Management (IAM) roles that are associated
      with the DB cluster. Each dictionary contains the role_arn and the status of the role.
  returned: always
  type: list
  sample: []
auto_minor_version_upgrade:
  description: Specifies whether minor engine upgrades are applied automatically to the DB cluster during the maintenance window.
  returned: always
  type: bool
  sample: true
availability_zones:
  description: The list of availability zones that instances in the DB cluster can be created in.
  returned: always
  type: list
  sample:
  - us-east-1c
  - us-east-1a
  - us-east-1e
backup_retention_period:
  description: The number of days for which automatic DB snapshots are retained.
  returned: always
  type: int
  sample: 1
changed:
  description: If the RDS cluster has changed.
  returned: always
  type: bool
  sample: true
cluster_create_time:
  description: The time in UTC when the DB cluster was created.
  returned: always
  type: str
  sample: "2018-06-29T14:08:58.491000+00:00"
copy_tags_to_snapshot:
  description:
    - Specifies whether tags are copied from the DB cluster to snapshots of the DB cluster.
  returned: always
  type: bool
  sample: false
cross_account_clone:
  description:
    - Specifies whether the DB cluster is a clone of a DB cluster owned by a different Amazon Web Services account.
  returned: always
  type: bool
  sample: false
db_cluster_arn:
  description: The Amazon Resource Name (ARN) for the DB cluster.
  returned: always
  type: str
  sample: arn:aws:rds:us-east-1:123456789012:cluster:rds-cluster-demo
db_cluster_identifier:
  description: The lowercase user-supplied DB cluster identifier.
  returned: always
  type: str
  sample: "rds-cluster-demo"
db_cluster_members:
  description:
    - A list of dictionaries containing information about the instances in the cluster.
      Each dictionary contains the db_instance_identifier, is_cluster_writer (bool),
      db_cluster_parameter_group_status, and promotion_tier (int).
  returned: always
  type: list
  sample: []
db_cluster_parameter_group:
  description: The parameter group associated with the DB cluster.
  returned: always
  type: str
  sample: "default.aurora5.6"
db_cluster_resource_id:
  description: The AWS Region-unique, immutable identifier for the DB cluster.
  returned: always
  type: str
  sample: "cluster-D2MEQDN3BQNXDF74K6DQJTHASU"
db_subnet_group:
  description: The name of the subnet group associated with the DB Cluster.
  returned: always
  type: str
  sample: "default"
deletion_protection:
  description:
    - Indicates if the DB cluster has deletion protection enabled.
      The database can't be deleted when deletion protection is enabled.
  returned: always
  type: bool
  sample: false
domain_memberships:
  description:
    - The Active Directory Domain membership records associated with the DB cluster.
  returned: always
  type: list
  sample: []
earliest_restorable_time:
  description: The earliest time to which a database can be restored with point-in-time restore.
  returned: always
  type: str
  sample: "2018-06-29T14:09:34.797000+00:00"
endpoint:
  description: The connection endpoint for the primary instance of the DB cluster.
  returned: always
  type: str
  sample: "rds-cluster-demo.cluster-cvlrtwiennww.us-east-1.rds.amazonaws.com"
engine:
  description: The database engine of the DB cluster.
  returned: always
  type: str
  sample: "aurora"
engine_mode:
  description: The DB engine mode of the DB cluster.
  returned: always
  type: str
  sample: "provisioned"
engine_version:
  description: The database engine version.
  returned: always
  type: str
  sample: "5.6.10a"
hosted_zone_id:
  description: The ID that Amazon Route 53 assigns when you create a hosted zone.
  returned: always
  type: str
  sample: "Z2R2ITUGPM61AM"
http_endpoint_enabled:
  description:
    - A value that indicates whether the HTTP endpoint for an Aurora Serverless DB cluster is enabled.
  returned: always
  type: bool
  sample: false
iam_database_authentication_enabled:
  description: Whether IAM accounts may be mapped to database accounts.
  returned: always
  type: bool
  sample: false
latest_restorable_time:
  description: The latest time to which a database can be restored with point-in-time restore.
  returned: always
  type: str
  sample: "2018-06-29T14:09:34.797000+00:00"
master_username:
  description: The master username for the DB cluster.
  returned: always
  type: str
  sample: "username"
multi_az:
  description: Whether the DB cluster has instances in multiple availability zones.
  returned: always
  type: bool
  sample: false
port:
  description: The port that the database engine is listening on.
  returned: always
  type: int
  sample: 3306
preferred_backup_window:
  description: The UTC weekly time range during which system maintenance can occur.
  returned: always
  type: str
  sample: "10:18-10:48"
preferred_maintenance_window:
  description: The UTC weekly time range during which system maintenance can occur.
  returned: always
  type: str
  sample: "tue:03:23-tue:03:53"
read_replica_identifiers:
  description: A list of read replica ID strings associated with the DB cluster.
  returned: always
  type: list
  sample: []
reader_endpoint:
  description: The reader endpoint for the DB cluster.
  returned: always
  type: str
  sample: "rds-cluster-demo.cluster-ro-cvlrtwiennww.us-east-1.rds.amazonaws.com"
serverless_v2_scaling_configuration:
  description: The scaling configuration for an Aurora Serverless v2 DB cluster.
  returned: when configured
  type: dict
  sample: {
      "max_capacity": 4.5,
      "min_capacity": 2.5
  }
  version_added: 7.3.0
status:
  description: The status of the DB cluster.
  returned: always
  type: str
  sample: "available"
storage_encrypted:
  description: Whether the DB cluster is storage encrypted.
  returned: always
  type: bool
  sample: false
tag_list:
  description: A list of tags consisting of key-value pairs.
  returned: always
  type: list
  elements: dict
  sample: [
    {
      "key": "Created_By",
      "value": "Ansible_rds_cluster_integration_test"
    }
  ]
tags:
  description: A dictionary of key value pairs.
  returned: always
  type: dict
  sample: {
    "Name": "rds-cluster-demo"
  }
vpc_security_groups:
  description: A list of the DB cluster's security groups and their status.
  returned: always
  type: complex
  contains:
    status:
      description: Status of the security group.
      returned: always
      type: str
      sample: "active"
    vpc_security_group_id:
      description: Security group of the cluster.
      returned: always
      type: str
      sample: "sg-12345678"
"""

try:
    import botocore
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.rds import arg_spec_to_rds_params
from ansible_collections.amazon.aws.plugins.module_utils.rds import call_method
from ansible_collections.amazon.aws.plugins.module_utils.rds import ensure_tags
from ansible_collections.amazon.aws.plugins.module_utils.rds import get_tags
from ansible_collections.amazon.aws.plugins.module_utils.rds import wait_for_cluster_status
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list


@AWSRetry.jittered_backoff(retries=10)
def _describe_db_clusters(**params):
    try:
        paginator = client.get_paginator("describe_db_clusters")
        return paginator.paginate(**params).build_full_result()["DBClusters"][0]
    except is_boto3_error_code("DBClusterNotFoundFault"):
        return {}


def get_add_role_options(params_dict, cluster):
    current_role_arns = [role["RoleArn"] for role in cluster.get("AssociatedRoles", [])]
    role = params_dict["RoleArn"]
    if role is not None and role not in current_role_arns:
        return {"RoleArn": role, "DBClusterIdentifier": params_dict["DBClusterIdentifier"]}
    return {}


def get_backtrack_options(params_dict):
    options = ["BacktrackTo", "DBClusterIdentifier", "UseEarliestTimeOnPointInTimeUnavailable"]
    if params_dict["BacktrackTo"] is not None:
        options = dict((k, params_dict[k]) for k in options if params_dict[k] is not None)
        if "ForceBacktrack" in params_dict:
            options["Force"] = params_dict["ForceBacktrack"]
        return options
    return {}


def get_create_options(params_dict):
    options = [
        "AvailabilityZones",
        "BacktrackWindow",
        "BackupRetentionPeriod",
        "PreferredBackupWindow",
        "CharacterSetName",
        "DBClusterIdentifier",
        "DBClusterParameterGroupName",
        "DBSubnetGroupName",
        "DatabaseName",
        "EnableCloudwatchLogsExports",
        "EnableIAMDatabaseAuthentication",
        "KmsKeyId",
        "Engine",
        "EngineMode",
        "EngineVersion",
        "PreferredMaintenanceWindow",
        "MasterUserPassword",
        "MasterUsername",
        "OptionGroupName",
        "Port",
        "ReplicationSourceIdentifier",
        "SourceRegion",
        "StorageEncrypted",
        "Tags",
        "VpcSecurityGroupIds",
        "EngineMode",
        "ScalingConfiguration",
        "DeletionProtection",
        "EnableHttpEndpoint",
        "CopyTagsToSnapshot",
        "Domain",
        "DomainIAMRoleName",
        "EnableGlobalWriteForwarding",
        "GlobalClusterIdentifier",
        "AllocatedStorage",
        "DBClusterInstanceClass",
        "StorageType",
        "Iops",
        "EngineMode",
        "ServerlessV2ScalingConfiguration",
    ]

    return dict((k, v) for k, v in params_dict.items() if k in options and v is not None)


def get_modify_options(params_dict, force_update_password):
    options = [
        "ApplyImmediately",
        "BacktrackWindow",
        "BackupRetentionPeriod",
        "PreferredBackupWindow",
        "DBClusterIdentifier",
        "DBClusterParameterGroupName",
        "EnableIAMDatabaseAuthentication",
        "EngineVersion",
        "PreferredMaintenanceWindow",
        "MasterUserPassword",
        "NewDBClusterIdentifier",
        "OptionGroupName",
        "Port",
        "VpcSecurityGroupIds",
        "EnableIAMDatabaseAuthentication",
        "CloudwatchLogsExportConfiguration",
        "DeletionProtection",
        "EnableHttpEndpoint",
        "CopyTagsToSnapshot",
        "EnableGlobalWriteForwarding",
        "Domain",
        "DomainIAMRoleName",
        "AllocatedStorage",
        "DBClusterInstanceClass",
        "StorageType",
        "Iops",
        "EngineMode",
        "ServerlessV2ScalingConfiguration",
    ]
    modify_options = dict((k, v) for k, v in params_dict.items() if k in options and v is not None)
    if not force_update_password:
        modify_options.pop("MasterUserPassword", None)
    return modify_options


def get_delete_options(params_dict):
    options = ["DBClusterIdentifier", "FinalSnapshotIdentifier", "SkipFinalSnapshot"]
    return dict((k, params_dict[k]) for k in options if params_dict[k] is not None)


def get_restore_s3_options(params_dict):
    options = [
        "AvailabilityZones",
        "BacktrackWindow",
        "BackupRetentionPeriod",
        "CharacterSetName",
        "DBClusterIdentifier",
        "DBClusterParameterGroupName",
        "DBSubnetGroupName",
        "DatabaseName",
        "EnableCloudwatchLogsExports",
        "EnableIAMDatabaseAuthentication",
        "Engine",
        "EngineVersion",
        "KmsKeyId",
        "MasterUserPassword",
        "MasterUsername",
        "OptionGroupName",
        "Port",
        "PreferredBackupWindow",
        "PreferredMaintenanceWindow",
        "S3BucketName",
        "S3IngestionRoleArn",
        "S3Prefix",
        "SourceEngine",
        "SourceEngineVersion",
        "StorageEncrypted",
        "Tags",
        "VpcSecurityGroupIds",
        "DeletionProtection",
        "EnableHttpEndpoint",
        "CopyTagsToSnapshot",
        "Domain",
        "DomainIAMRoleName",
    ]

    return dict((k, v) for k, v in params_dict.items() if k in options and v is not None)


def get_restore_snapshot_options(params_dict):
    options = [
        "AvailabilityZones",
        "BacktrackWindow",
        "DBClusterIdentifier",
        "DBSubnetGroupName",
        "DatabaseName",
        "EnableCloudwatchLogsExports",
        "EnableIAMDatabaseAuthentication",
        "Engine",
        "EngineVersion",
        "KmsKeyId",
        "OptionGroupName",
        "Port",
        "SnapshotIdentifier",
        "Tags",
        "VpcSecurityGroupIds",
        "DBClusterParameterGroupName",
        "DeletionProtection",
        "CopyTagsToSnapshot",
        "Domain",
        "DomainIAMRoleName",
    ]
    return dict((k, v) for k, v in params_dict.items() if k in options and v is not None)


def get_restore_cluster_options(params_dict):
    options = [
        "BacktrackWindow",
        "DBClusterIdentifier",
        "DBSubnetGroupName",
        "EnableCloudwatchLogsExports",
        "EnableIAMDatabaseAuthentication",
        "KmsKeyId",
        "OptionGroupName",
        "Port",
        "RestoreToTime",
        "RestoreType",
        "SourceDBClusterIdentifier",
        "Tags",
        "UseLatestRestorableTime",
        "VpcSecurityGroupIds",
        "DeletionProtection",
        "CopyTagsToSnapshot",
        "Domain",
        "DomainIAMRoleName",
    ]
    return dict((k, v) for k, v in params_dict.items() if k in options and v is not None)


def get_rds_method_attribute_name(cluster):
    state = module.params["state"]
    creation_source = module.params["creation_source"]
    method_name = None
    method_options_name = None

    if state == "absent":
        if cluster and cluster["Status"] not in ["deleting", "deleted"]:
            method_name = "delete_db_cluster"
            method_options_name = "get_delete_options"
    elif state == "started":
        if cluster and cluster["Status"] not in ["starting", "started", "available"]:
            method_name = "start_db_cluster"
            method_options_name = "get_modify_options"
    elif state == "stopped":
        if cluster and cluster["Status"] not in ["stopping", "stopped"]:
            method_name = "stop_db_cluster"
            method_options_name = "get_modify_options"
    else:
        if cluster:
            method_name = "modify_db_cluster"
            method_options_name = "get_modify_options"
        elif creation_source == "snapshot":
            method_name = "restore_db_cluster_from_snapshot"
            method_options_name = "get_restore_snapshot_options"
        elif creation_source == "s3":
            method_name = "restore_db_cluster_from_s3"
            method_options_name = "get_restore_s3_options"
        elif creation_source == "cluster":
            method_name = "restore_db_cluster_to_point_in_time"
            method_options_name = "get_restore_cluster_options"
        else:
            method_name = "create_db_cluster"
            method_options_name = "get_create_options"

    return method_name, method_options_name


def add_role(params):
    if not module.check_mode:
        try:
            client.add_role_to_db_cluster(**params)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            module.fail_json_aws(
                e, msg=f"Unable to add role {params['RoleArn']} to cluster {params['DBClusterIdentifier']}"
            )
        wait_for_cluster_status(client, module, params["DBClusterIdentifier"], "cluster_available")


def backtrack_cluster(params):
    if not module.check_mode:
        try:
            client.backtrack_db_cluster(**params)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            module.fail_json_aws(e, msg=f"Unable to backtrack cluster {params['DBClusterIdentifier']}")
        wait_for_cluster_status(client, module, params["DBClusterIdentifier"], "cluster_available")


def get_cluster(db_cluster_id):
    try:
        return _describe_db_clusters(DBClusterIdentifier=db_cluster_id)
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg="Failed to describe DB clusters")


def changing_cluster_options(modify_params, current_cluster):
    changing_params = {}
    apply_immediately = modify_params.pop("ApplyImmediately")
    db_cluster_id = modify_params.pop("DBClusterIdentifier")

    enable_cloudwatch_logs_export = modify_params.pop("EnableCloudwatchLogsExports", None)
    if enable_cloudwatch_logs_export is not None:
        desired_cloudwatch_logs_configuration = {"EnableLogTypes": [], "DisableLogTypes": []}
        provided_cloudwatch_logs = set(enable_cloudwatch_logs_export)
        current_cloudwatch_logs_export = set(current_cluster["EnabledCloudwatchLogsExports"])

        desired_cloudwatch_logs_configuration["EnableLogTypes"] = sorted(
            list(provided_cloudwatch_logs.difference(current_cloudwatch_logs_export))
        )
        if module.params["purge_cloudwatch_logs_exports"]:
            desired_cloudwatch_logs_configuration["DisableLogTypes"] = sorted(
                list(current_cloudwatch_logs_export.difference(provided_cloudwatch_logs))
            )
        if (
            desired_cloudwatch_logs_configuration["EnableLogTypes"]
            or desired_cloudwatch_logs_configuration["DisableLogTypes"]
        ):
            changing_params["CloudwatchLogsExportConfiguration"] = desired_cloudwatch_logs_configuration

    password = modify_params.pop("MasterUserPassword", None)
    if password:
        changing_params["MasterUserPassword"] = password

    new_cluster_id = modify_params.pop("NewDBClusterIdentifier", None)
    if new_cluster_id and new_cluster_id != current_cluster["DBClusterIdentifier"]:
        changing_params["NewDBClusterIdentifier"] = new_cluster_id

    option_group = modify_params.pop("OptionGroupName", None)
    if option_group and option_group not in [
        g["DBClusterOptionGroupName"] for g in current_cluster["DBClusterOptionGroupMemberships"]
    ]:
        changing_params["OptionGroupName"] = option_group

    vpc_sgs = modify_params.pop("VpcSecurityGroupIds", None)
    if vpc_sgs:
        desired_vpc_sgs = []
        provided_vpc_sgs = set(vpc_sgs)
        current_vpc_sgs = set([sg["VpcSecurityGroupId"] for sg in current_cluster["VpcSecurityGroups"]])
        if module.params["purge_security_groups"]:
            desired_vpc_sgs = sorted(list(provided_vpc_sgs))
        else:
            if provided_vpc_sgs - current_vpc_sgs:
                desired_vpc_sgs = sorted(list(provided_vpc_sgs | current_vpc_sgs))

        if desired_vpc_sgs:
            changing_params["VpcSecurityGroupIds"] = desired_vpc_sgs

    desired_db_cluster_parameter_group = modify_params.pop("DBClusterParameterGroupName", None)
    if desired_db_cluster_parameter_group:
        if desired_db_cluster_parameter_group != current_cluster["DBClusterParameterGroup"]:
            changing_params["DBClusterParameterGroupName"] = desired_db_cluster_parameter_group

    for param in modify_params:
        # describe_db_clusters does not return StorageType for clusters with storage config as "aurora standard"
        if param == "StorageType":
            changing_params[param] = modify_params[param]
            continue
        # describe_db_clusters returns the Domain member ship information as follow
        # 'DomainMemberships': [
        #     {
        #         'Domain': 'string',
        #         'Status': 'string',
        #         'FQDN': 'string',
        #         'IAMRoleName': 'string'
        #     },
        # ]
        if param in ("Domain", "DomainIAMRoleName"):
            n_domain = modify_params.get("Domain")
            n_domain_iam_role_name = modify_params.get("DomainIAMRoleName")
            if not any(
                n_domain == d.get("Domain") and n_domain_iam_role_name == d.get("IAMRoleName")
                for d in current_cluster.get("DomainMemberships", [])
            ):
                changing_params[param] = modify_params[param]
            continue
        if modify_params[param] != current_cluster[param]:
            changing_params[param] = modify_params[param]

    if changing_params:
        changing_params["DBClusterIdentifier"] = db_cluster_id
        if apply_immediately is not None:
            changing_params["ApplyImmediately"] = apply_immediately

    if module.params["state"] == "started":
        if current_cluster["Engine"] in ["mysql", "postgres"]:
            module.fail_json("Only aurora clusters can use the state started")
        # start_db_cluster supports only DBClusterIdentifier
        changing_params = {}
        changing_params["DBClusterIdentifier"] = db_cluster_id

    if module.params["state"] == "stopped":
        if current_cluster["Engine"] in ["mysql", "postgres"]:
            module.fail_json("Only aurora clusters can use the state stopped")
        # stop_db_cluster supports only DBClusterIdentifier
        changing_params = {}
        changing_params["DBClusterIdentifier"] = db_cluster_id

    return changing_params


def ensure_present(cluster, parameters, method_name, method_options_name):
    changed = False

    if not cluster:
        if parameters.get("Tags") is not None:
            parameters["Tags"] = ansible_dict_to_boto3_tag_list(parameters["Tags"])

        call_method(client, module, method_name, eval(method_options_name)(parameters))
        changed = True
    else:
        if get_backtrack_options(parameters):
            backtrack_cluster(client, module, get_backtrack_options(parameters))
            changed = True
        else:
            modifiable_options = eval(method_options_name)(
                parameters, force_update_password=module.params["force_update_password"]
            )
            modify_options = changing_cluster_options(modifiable_options, cluster)
            if modify_options:
                call_method(client, module, method_name, modify_options)
                changed = True
            if module.params["tags"] is not None:
                existing_tags = get_tags(client, module, cluster["DBClusterArn"])
                changed |= ensure_tags(
                    client,
                    module,
                    cluster["DBClusterArn"],
                    existing_tags,
                    module.params["tags"],
                    module.params["purge_tags"],
                )

    add_role_params = get_add_role_options(parameters, cluster)
    if add_role_params:
        add_role(client, module, add_role_params)
        changed = True

    if module.params["promote"] and cluster.get("ReplicationSourceIdentifier"):
        call_method(
            client,
            module,
            "promote_read_replica_db_cluster",
            parameters={"DBClusterIdentifier": module.params["db_cluster_identifier"]},
        )
        changed = True

    return changed


def handle_remove_from_global_db(cluster):
    global_cluster_id = module.params.get("global_cluster_identifier")
    db_cluster_id = module.params.get("db_cluster_identifier")
    db_cluster_arn = cluster["DBClusterArn"]

    if module.check_mode:
        return True

    try:
        client.remove_from_global_cluster(DbClusterIdentifier=db_cluster_arn, GlobalClusterIdentifier=global_cluster_id)
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(
            e, msg=f"Failed to remove cluster {db_cluster_id} from global DB cluster {global_cluster_id}."
        )

    # for replica cluster - wait for cluster to change status from 'available' to 'promoting'
    # only replica/secondary clusters have "GlobalWriteForwardingStatus" field
    if "GlobalWriteForwardingStatus" in cluster:
        wait_for_cluster_status(client, module, db_cluster_id, "db_cluster_promoting")

    # if wait=true, wait for db cluster remove from global db operation to complete
    if module.params.get("wait"):
        wait_for_cluster_status(client, module, db_cluster_id, "cluster_available")

    return True


def main():
    global module
    global client

    arg_spec = dict(
        state=dict(choices=["present", "absent", "started", "stopped"], default="present"),
        creation_source=dict(type="str", choices=["snapshot", "s3", "cluster"]),
        force_update_password=dict(type="bool", default=False),
        promote=dict(type="bool", default=False),
        purge_cloudwatch_logs_exports=dict(type="bool", default=True),
        purge_tags=dict(type="bool", default=True),
        wait=dict(type="bool", default=True),
        purge_security_groups=dict(type="bool", default=True),
    )

    parameter_options = dict(
        apply_immediately=dict(type="bool", default=False),
        availability_zones=dict(type="list", elements="str", aliases=["zones", "az"]),
        backtrack_to=dict(),
        backtrack_window=dict(type="int"),
        backup_retention_period=dict(type="int", default=1),
        character_set_name=dict(),
        database_name=dict(aliases=["db_name"]),
        db_cluster_identifier=dict(required=True, aliases=["cluster_id", "id", "cluster_name"]),
        db_cluster_parameter_group_name=dict(),
        db_subnet_group_name=dict(),
        enable_cloudwatch_logs_exports=dict(type="list", elements="str"),
        deletion_protection=dict(type="bool"),
        global_cluster_identifier=dict(),
        enable_http_endpoint=dict(type="bool"),
        copy_tags_to_snapshot=dict(type="bool"),
        domain=dict(),
        domain_iam_role_name=dict(),
        enable_global_write_forwarding=dict(type="bool"),
        db_cluster_instance_class=dict(type="str"),
        enable_iam_database_authentication=dict(type="bool"),
        engine=dict(choices=["aurora", "aurora-mysql", "aurora-postgresql", "mysql", "postgres"]),
        engine_mode=dict(choices=["provisioned", "serverless", "parallelquery", "global", "multimaster"]),
        engine_version=dict(),
        allocated_storage=dict(type="int"),
        storage_type=dict(type="str", choices=["io1", "aurora", "aurora-iopt1"]),
        iops=dict(type="int"),
        final_snapshot_identifier=dict(),
        force_backtrack=dict(type="bool"),
        kms_key_id=dict(),
        master_user_password=dict(aliases=["password"], no_log=True),
        master_username=dict(aliases=["username"]),
        new_db_cluster_identifier=dict(aliases=["new_cluster_id", "new_id", "new_cluster_name"]),
        option_group_name=dict(),
        port=dict(type="int"),
        preferred_backup_window=dict(aliases=["backup_window"]),
        preferred_maintenance_window=dict(aliases=["maintenance_window"]),
        remove_from_global_db=dict(type="bool"),
        replication_source_identifier=dict(aliases=["replication_src_id"]),
        restore_to_time=dict(),
        restore_type=dict(choices=["full-copy", "copy-on-write"]),
        role_arn=dict(),
        s3_bucket_name=dict(),
        s3_ingestion_role_arn=dict(),
        s3_prefix=dict(),
        serverless_v2_scaling_configuration=dict(
            type="dict",
            options=dict(
                min_capacity=dict(type="float"),
                max_capacity=dict(type="float"),
            ),
        ),
        skip_final_snapshot=dict(type="bool", default=False),
        snapshot_identifier=dict(),
        source_db_cluster_identifier=dict(),
        source_engine=dict(choices=["mysql"]),
        source_engine_version=dict(),
        source_region=dict(),
        storage_encrypted=dict(type="bool"),
        tags=dict(type="dict", aliases=["resource_tags"]),
        use_earliest_time_on_point_in_time_unavailable=dict(type="bool"),
        use_latest_restorable_time=dict(type="bool"),
        vpc_security_group_ids=dict(type="list", elements="str"),
    )
    arg_spec.update(parameter_options)

    required_by_s3_creation_source = [
        "s3_bucket_name",
        "engine",
        "master_username",
        "master_user_password",
        "source_engine",
        "source_engine_version",
        "s3_ingestion_role_arn",
    ]

    module = AnsibleAWSModule(
        argument_spec=arg_spec,
        required_if=[
            ["creation_source", "snapshot", ["snapshot_identifier", "engine"]],
            ["creation_source", "s3", required_by_s3_creation_source],
            ["remove_from_global_db", True, ["global_cluster_identifier", "db_cluster_identifier"]],
        ],
        mutually_exclusive=[
            ["s3_bucket_name", "source_db_cluster_identifier", "snapshot_identifier"],
            ["use_latest_restorable_time", "restore_to_time"],
        ],
        supports_check_mode=True,
    )

    retry_decorator = AWSRetry.jittered_backoff(retries=10)

    try:
        client = module.client("rds", retry_decorator=retry_decorator)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS.")

    if module.params.get("engine") and module.params["engine"] in ("mysql", "postgres"):
        if module.params["state"] == "present":
            if not (
                module.params.get("allocated_storage")
                and module.params.get("iops")
                and module.params.get("db_cluster_instance_class")
            ):
                module.fail_json(
                    f"When engine={module.params['engine']} allocated_storage, iops and db_cluster_instance_class must be specified"
                )
            else:
                # Fall to default value
                if not module.params.get("storage_type"):
                    module.params["storage_type"] = "io1"

    if module.params.get("engine") and module.params["engine"] in ("aurora", "aurora-mysql", "aurora-postgresql"):
        if not module.check_mode and module.params.get("storage_type"):
            module.params["storage_type"] = "aurora"

    if module.params.get("storage_type") and module.params["storage_type"] in ("aurora", "aurora-iopt1"):
        if module.params.get("engine") not in ("aurora", "aurora-mysql", "aurora-postgresql"):
            module.fail_json(
                f"storage_type={module.params['storage_type']} is only supported for aurora engines 'aurora', 'aurora-mysql', 'aurora-postgresql'"
            )

    module.params["db_cluster_identifier"] = module.params["db_cluster_identifier"].lower()
    cluster = get_cluster(module.params["db_cluster_identifier"])

    if module.params["new_db_cluster_identifier"]:
        module.params["new_db_cluster_identifier"] = module.params["new_db_cluster_identifier"].lower()

        if get_cluster(module.params["new_db_cluster_identifier"]):
            module.fail_json(
                f"A new cluster ID {module.params['new_db_cluster_identifier']} was provided but it already exists"
            )
        if not cluster:
            module.fail_json(
                f"A new cluster ID {module.params['new_db_cluster_identifier']} was provided but the cluster to be renamed does not exist"
            )

    if (
        module.params["state"] == "absent"
        and module.params["skip_final_snapshot"] is False
        and module.params["final_snapshot_identifier"] is None
    ):
        module.fail_json(
            msg="skip_final_snapshot is False but all of the following are missing: final_snapshot_identifier"
        )

    changed = False

    parameters = arg_spec_to_rds_params(dict((k, module.params[k]) for k in module.params if k in parameter_options))
    method_name, method_options_name = get_rds_method_attribute_name(cluster)

    if method_name:
        if method_name == "delete_db_cluster":
            if cluster and module.params.get("remove_from_global_db"):
                if cluster["Engine"] in ["aurora", "aurora-mysql", "aurora-postgresql"]:
                    changed = handle_remove_from_global_db(cluster)

            call_method(client, module, method_name, eval(method_options_name)(parameters))
            changed = True
        else:
            changed |= ensure_present(cluster, parameters, method_name, method_options_name)

    if not module.check_mode and module.params["new_db_cluster_identifier"] and module.params["apply_immediately"]:
        cluster_id = module.params["new_db_cluster_identifier"]
    else:
        cluster_id = module.params["db_cluster_identifier"]

    if cluster_id and get_cluster(cluster_id) and module.params.get("remove_from_global_db"):
        if cluster["Engine"] in ["aurora", "aurora-mysql", "aurora-postgresql"]:
            if changed:
                wait_for_cluster_status(client, module, cluster_id, "cluster_available")
        changed |= handle_remove_from_global_db(cluster)

    result = camel_dict_to_snake_dict(get_cluster(cluster_id))

    if result:
        result["tags"] = get_tags(client, module, result["db_cluster_arn"])

    module.exit_json(changed=changed, **result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: ucs_scrub_policy

short_description: Manages UCS Scrub Policies on UCS Manager

description:
  - Manages UCS Scrub Policies on UCS Manager.

extends_documentation_fragment: cisco.ucs.ucs

options:
    state:
        description:
        - If C(absent), will remove organization.
        - If C(present), will create or update organization.
        choices: [absent, present]
        default: present
        type: str

    name:
        description:
        - The name of the organization.
        - Enter up to 16 characters.
        - "You can use any characters or spaces except the following:"
        - "` (accent mark), \ (backslash), ^ (carat), \" (double quote)"
        - "= (equal sign), > (greater than), < (less than), ' (single quote)."
        required: true
        type: str

    description:
        description:
        - A user-defined description of the organization.
        - Enter up to 256 characters.
        - "You can use any characters or spaces except the following:"
        - "` (accent mark), \ (backslash), ^ (carat), \" (double quote)"
        - "= (equal sign), > (greater than), < (less than), ' (single quote)."
        aliases: [ descr ]
        type: str

    bios_settings_scrub:
        description:
        - Scrub the BIOS settings.
        - If the field is set to Yes, when a service profile containing this
        - scrub policy is disassociated from a server, the BIOS settings for
        - that server are erased and reset to the defaults for that server
        - type and vendor. If this field is set to No, the BIOS settings are
        - preserved.
        - yes scrub the BIOS settings.
        - no do not scrub the BIOS settings.
        choices: ['yes', 'no']
        type: str

    disk_scrub:
        description:
        - Scrub the BIOS settings.
        - If this field is set to Yes, when a service profile containing this
        - scrub policy is disassociated from a server, all data on the server
        - local drives is completely erased. If this field is set to No, the
        - data on the local drives is preserved, including all local storage
        - configuration.
        - yes scrub the server disks.
        - no do not scrub the server disks.
        choices: ['yes', 'no']
        type: str

    flex_flash_scrub:
        description:
        - Scrub the BIOS settings.
        - If the field is set to Yes, the HV partition on the SD card is
        - formatted using the PNUOS formatting utility when the server is
        - reacknowledged. If this field is set to No, the SD card is preserved.
        - yes scrub the flex flash.
        - no do not scrub the flex flash.
        choices: ['yes', 'no']
        type: str

    persistent_memory_scrub:
        description:
        - Scrub the BIOS settings.
        - If the field is set to Yes, when a service profile containing this
        - scrub policy is disassociated from a server, all persistent memory
        - modules for that server are erased and reset to the defaults for that
        - server type and vendor. If this field is set to No, the persistent
        - memory modules are preserved.
        - yes scrub the persistent memory.
        - no do not scrub the persistent memory.
        choices: ['yes', 'no']
        type: str

    org_dn:
        description:
        - Org dn (distinguished name)
        default: org-root
        type: str

requirements:
- ucsmsdk

author:
- John McDonough (@movinalot)
'''

EXAMPLES = r'''
- name: Add UCS Scrub Policy
  cisco.ucs.ucs_scrub_policy:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    state: present
    description: Scrub All Policy
    name: all_scrub
    bios_settings_scrub: yes
    disk_scrub: yes
    flex_flash_scrub: yes
    persistent_memory_scrub: yes
  delegate_to: localhost

- name: Add UCS Scrub Policy in an Organization
  cisco.ucs.ucs_scrub_policy:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    state: present
    org_dn: org-root/org-prod
    name: all_scrub
    description: Scrub All Policy Org Prod servers
    bios_settings_scrub: yes
    disk_scrub: yes
    flex_flash_scrub: yes
    persistent_memory_scrub: yes
  delegate_to: localhost

- name: Update UCS Scrub Policy
  cisco.ucs.ucs_scrub_policy:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    state: present
    org_dn: org-root/org-prod
    name: BD_scrub
    description: Scrub BIOS and Disk Policy Org Prod servers
    bios_settings_scrub: yes
    disk_scrub: yes
    flex_flash_scrub: no
    persistent_memory_scrub: no
  delegate_to: localhost

- name: Update UCS Scrub Policy
  cisco.ucs.ucs_scrub_policy:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    state: present
    org_dn: org-root/org-prod
    name: BD_scrub
    description: Scrub BIOS and Disk Policy Org Prod servers
    bios_settings_scrub: yes
    disk_scrub: yes
    flex_flash_scrub: yes
  delegate_to: localhost

- name: Delete UCS Scrub Policy
  cisco.ucs.ucs_scrub_policy:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    state: absent
    org_dn: org-root/org-prod
    name: BD_scrub
  delegate_to: localhost

- name: Delete UCS Scrub Policy
  cisco.ucs.ucs_scrub_policy:
    hostname: "{{ ucs_hostname }}"
    username: "{{ ucs_username }}"
    password: "{{ ucs_password }}"
    state: absent
    name: BD_scrub
  delegate_to: localhost
'''

RETURN = r'''
#
'''

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.ucs.plugins.module_utils.ucs import (
    UCSModule,
    ucs_argument_spec
)


def main():
    argument_spec = ucs_argument_spec.copy()
    argument_spec.update(
        org_dn=dict(type='str', default='org-root'),
        name=dict(required=True, type='str'),
        description=dict(type='str', aliases=['descr']),
        bios_settings_scrub=dict(type='str', choices=['yes', 'no']),
        disk_scrub=dict(type='str', choices=['yes', 'no']),
        flex_flash_scrub=dict(type='str', choices=['yes', 'no']),
        persistent_memory_scrub=dict(type='str', choices=['yes', 'no']),
        state=dict(
            type='str', default='present',
            choices=['present', 'absent']
        ),
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
        required_if=[
            ['state', 'present', ['name']],
        ],
    )

    # UCSModule verifies ucsmsdk is present and exits on failure.
    # Imports are below for UCS object creation.
    ucs = UCSModule(module)
    from importlib import import_module
    from ucsmsdk.ucscoreutils import get_meta_info

    # The Class(es) this module is managing
    module_file = 'ucsmsdk.mometa.compute.ComputeScrubPolicy'
    module_class = 'ComputeScrubPolicy'
    mo_module = import_module(module_file)
    mo_class = getattr(mo_module, module_class)

    META = get_meta_info(class_id=module_class)

    err = False
    changed = False
    requested_state = module.params['state']

    kwargs = dict()

    # Manage Attributes
    for attribute in [
            'bios_settings_scrub',
            'disk_scrub',
            'flex_flash_scrub',
            'persistent_memory_scrub'
    ]:
        if module.params[attribute] is not None:
            kwargs[attribute] = module.params[attribute]

    kwargs['descr'] = module.params['description']

    try:
        dn = (
            module.params['org_dn'] + '/' +
            META.rn[0:META.rn.index('-') + 1] +
            module.params['name']
        )
        mo = ucs.login_handle.query_dn(dn)

        # Determine state change
        if mo:
            # Object exists, if it should exist has anything changed?
            if requested_state == 'present':
                # Do some or all Object properties not match, that is a change

                if not mo.check_prop_match(**kwargs):
                    changed = True

        # Object does not exist but should, that is a change
        else:
            if requested_state == 'present':
                changed = True

        # Object exists but should not, that is a change
        if mo and requested_state == 'absent':
            changed = True

        # Apply state if not check_mode
        if changed and not module.check_mode:
            if requested_state == 'absent':
                ucs.login_handle.remove_mo(mo)
            else:
                kwargs['parent_mo_or_dn'] = module.params['org_dn']
                kwargs['name'] = module.params['name']

                mo = mo_class(**kwargs)
                ucs.login_handle.add_mo(mo, modify_present=True)
            ucs.login_handle.commit()

    except Exception as e:
        err = True
        ucs.result['msg'] = "setup error: %s " % str(e)

    ucs.result['changed'] = changed
    if err:
        module.fail_json(**ucs.result)

    module.exit_json(**ucs.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2018 Yunge Zhu, <yungez@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_webapp
version_added: "0.1.2"
short_description: Manage Web App instances
description:
    - Create, update and delete instance of Web App.

options:
    resource_group:
        description:
            - Name of the resource group to which the resource belongs.
        required: True
        type: str
    name:
        description:
            - Unique name of the app to create or update. To create or update a deployment slot, use the {slot} parameter.
        required: True
        type: str

    location:
        description:
            - Resource location. If not set, location from the resource group will be used as default.
        type: str

    plan:
        description:
            - App service plan. Required for creation.
            - Can be name of existing app service plan in same resource group as web app.
            - Can be the resource ID of an existing app service plan. For example
              /subscriptions/<subs_id>/resourceGroups/<resource_group>/providers/Microsoft.Web/serverFarms/<plan_name>.
            - Can be a dict containing five parameters, defined below.
            - C(name), name of app service plan.
            - C(resource_group), resource group of the app service plan.
            - C(sku), SKU of app service plan, allowed values listed on U(https://azure.microsoft.com/en-us/pricing/details/app-service/linux/).
            - C(is_linux), whether or not the app service plan is Linux. defaults to C(False).
            - C(number_of_workers), number of workers for app service plan.
        type: raw

    frameworks:
        description:
            - Set of run time framework settings. Each setting is a dictionary.
            - See U(https://docs.microsoft.com/en-us/azure/app-service/app-service-web-overview) for more info.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - Name of the framework.
                    - Supported framework list for Windows web app and Linux web app is different.
                    - Windows web apps support C(java), C(net_framework), C(php), C(python), and C(node) from June 2018.
                    - Windows web apps support multiple framework at the same time.
                    - Linux web apps support C(java), C(ruby), C(php), C(python), C(dotnetcore), and C(node) from June 2018.
                    - Linux web apps support only one framework.
                    - Java framework is mutually exclusive with others.
                type: str
                required: true
                choices:
                    - java
                    - net_framework
                    - php
                    - python
                    - ruby
                    - dotnetcore
                    - node
            version:
                description:
                    - Version of the framework. For Linux web app supported value, see U(https://aka.ms/linux-stacks) for more info.
                    - C(net_framework) supported value sample, C(v4.8) for .NET 4.8 and C(v3.5) for .NET 3.5.
                    - C(php) supported value sample, C(8.1), C(8.2).
                    - C(python) supported value sample, C(3.8), C(3.9), C(3.10), C(3.11), C(3.12).
                    - C(node) supported value sample, C(18), C(20).
                    - C(dotnetcore) supported value sample, C(8), C(7), C(6).
                    - C(ruby) supported value sample, C(2.3).
                    - C(java) supported value sample, C(21), C(17), C(11) and C(8).
                type: str
                required: true
            settings:
                description:
                    - List of settings of the framework.
                type: dict
                suboptions:
                    java_container:
                        description:
                            - Name of Java container.
                            - Supported only when I(frameworks=java). Sample values C(Tomcat), C(JavaSE), C(RedHat).
                        type: str
                        required: True
                    java_container_version:
                        description:
                            - Version of Java container.
                            - Supported only when I(frameworks=java).
                            - Sample values for C(Tomcat), C(8.5), C(9.0), C(10.0), C(10.1).
                        type: str
                        required: True

    container_settings:
        description:
            - Web app container settings.
        type: dict
        suboptions:
            name:
                description:
                    - Name of the container, for example C(imagename:tag).
                    - To create a multi-container app, the name should be 'COMPOSE|' or 'KUBE|' followed by base64 encoded configuration.
                type: str
                required: True
            registry_server_url:
                description:
                    - Container registry server URL, for example C(mydockerregistry.io).
                type: str
            registry_server_user:
                description:
                    - The container registry server user name.
                type: str
            registry_server_password:
                description:
                    - The container registry server password.
                type: str

    scm_type:
        description:
            - Repository type of deployment source, for example C(LocalGit), C(GitHub).
            - List of supported values maintained at U(https://docs.microsoft.com/en-us/rest/api/appservice/webapps/createorupdate#scmtype).
        type: str

    always_on:
        description:
            - Keeps the app loaded even when there's no traffic.
        type: bool

    http20_enabled:
        description:
            - Configures a web site to allow clients to connect over HTTP 2.0.
        type: bool

    min_tls_version:
        description:
            - The minimum TLS encryption version required for the app.
        type: str
        choices:
            - '1.0'
            - '1.1'
            - '1.2'

    ftps_state:
        description:
            - The state of the FTP/FTPS service.
        type: str
        choices:
            - AllAllowed
            - FtpsOnly
            - Disabled

    deployment_source:
        description:
            - Deployment source for git.
        type: dict
        suboptions:
            url:
                description:
                    - Repository url of deployment source.
                type: str

            branch:
                description:
                    - The branch name of the repository.
                type: str
    startup_file:
        description:
            - The web's startup file.
            - Used only for Linux web apps.
        type: str

    client_affinity_enabled:
        description:
            - Whether or not to send session affinity cookies, which route client requests in the same session to the same instance.
        type: bool
        default: True

    https_only:
        description:
            - Configures web site to accept only https requests.
        type: bool

    app_settings:
        description:
            - Configure web app application settings. Suboptions are in key value pair format.
        type: dict

    purge_app_settings:
        description:
            - Purge any existing application settings. Replace web app application settings with app_settings.
        type: bool
        default: False

    app_state:
        description:
            - Start/Stop/Restart the web app.
        type: str
        choices:
            - started
            - stopped
            - restarted
        default: started
    site_auth_settings:
        description:
            - Configuration settings for the Azure App Service Authentication / Authorization feature.
        type: dict
        suboptions:
            kind:
                description:
                    - Kind of resource.
                type: str
            enabled:
                description:
                    - Whether enable or disable the Authentication / Authorization feature for the current app.
                type: bool
            runtime_version:
                description:
                    - The RuntimeVersion of the Authentication / Authorization feature in use for the current app.
                type: str
            unauthenticated_client_action:
                description:
                    - The action to take when an unauthenticated client attempts to access the app.
                type: str
                choices:
                    - RedirectToLoginPage
                    - AllowAnonymous
            token_store_enabled:
                description:
                    - Whether to use App Service Token Store.
                type: bool
            allowed_external_redirect_urls:
                description:
                    - External URLs that can be redirected to as part of logging in or logging out of the app.
                    - Note that the query string part of the URL is ignored.
                type: list
                elements: str
            default_provider:
                description:
                    - The default authentication provider to use when multiple providers are configured.
                type: str
                choices:
                    - AzureActiveDirectory
                    - Facebook
                    - Google
                    - MicrosoftAccount
                    - Twitter
                    - Github
            token_refresh_extension_hours:
                description:
                    - The number of hours after session token expiration that a session token can be used to call the token refresh API.
                type: float
            client_id:
                description:
                    - The Client ID of this relying party application, known as the client_id.
                type: str
            client_secret:
                description:
                    - The Client Secret of this relying party application (in Azure Active Directory, this is also referred to as the Key).
                type: str
            client_secret_setting_name:
                description:
                    - The app setting name that contains the client secret of the relying party application.
                type: str
            client_secret_certificate_thumbprint:
                description:
                    - An alternative to the client secret, that is the thumbprint of a certificate used for signing purposes.
                    - This property acts as a replacement for the Client Secret. It is also optional.
                type: str
            issuer:
                description:
                    - The OpenID Connect Issuer URI that represents the entity which issues access tokens for this application.
                type: str
            allowed_audiences:
                description:
                    - Allowed audience values to consider when validating JWTs issued by Azure Active Directory.
                type: list
                elements: str
            aad_claims_authorization:
                description:
                    - Gets a JSON string containing the Azure AD Acl settings.
                type: str
            additional_login_params:
                description:
                    - Login parameters to send to the OpenID Connect authorization endpoint when a user logs in.
                    - Each parameter must be in the form "key=value".
                type: str
            google_client_id:
                description:
                    - The OpenID Connect Client ID for the Google web application.
                type: str
            google_client_secret:
                description:
                    - The client secret associated with the Google web application.
                type: str
            google_client_secret_setting_name:
                description:
                    - The app setting name that contains the client secret associated with the Google web application.
                type: str
            google_o_auth_scopes:
                description:
                    - The OAuth 2.0 scopes that will be requested as part of Google Sign-In authentication.
                    - This setting is optional. If not specified, "openid", "profile", and "email" are used as default scopes.
                type: list
                elements: str
            facebook_app_id:
                description:
                    - The App ID of the Facebook app used for login.
                type: str
            facebook_app_secret:
                description:
                    - The App Secret of the Facebook app used for Facebook Login.
                type: str
            facebook_app_secret_setting_name:
                description:
                    - The app setting name that contains the app secret used for Facebook Login.
                type: str
            facebook_o_auth_scopes:
                description:
                    - The OAuth 2.0 scopes that will be requested as part of Facebook for Facebook Login.
                type: list
                elements: str
            git_hub_client_id:
                description:
                    - The Client Id of the GitHub app used for login.
                type: str
            git_hub_client_secret:
                description:
                    - The Client Secret of the GitHub app used for Github Login.
                type: str
            git_hub_client_secret_setting_name:
                description:
                    - The app setting name that contains the client secret of the Github app used for GitHub Login.
                type: str
            git_hub_o_auth_scopes:
                description:
                    - The OAuth 2.0 scopes that will be requested as part of GitHub Login authentication.
                    - This setting is optional.
                type: list
                elements: str
            twitter_consumer_key:
                description:
                    - The OAuth 1.0a consumer key of the Twitter application used for sign-in.
                type: str
            twitter_consumer_secret:
                description:
                    - The OAuth 1.0a consumer secret of the Twitter application used for sign-in.
                    - This setting is required for enabling Twitter Sign-In.
                type: str
            twitter_consumer_secret_setting_name:
                description:
                    - The app setting name that contains the OAuth 1.0a consumer secret of the Twitter application used for sign-in.
                type: str
            microsoft_account_client_id:
                description:
                    - The OAuth 2.0 client ID that was created for the app used for authentication.
                    - This setting is required for enabling Microsoft Account authentication.
                type: str
            microsoft_account_client_secret:
                description:
                    - The OAuth 2.0 client secret that was created for the app used for authentication.
                type: str
            microsoft_account_client_secret_setting_name:
                description:
                    - The app setting name containing the OAuth 2.0 client secret that was created for the app used for authentication.
                type: str
            microsoft_account_o_auth_scopes:
                description:
                    - The OAuth 2.0 scopes that will be requested as part of Microsoft Account authentication.
                type: list
                elements: str
            is_auth_from_file:
                description:
                    - If I(is_auth_from_file=true), the auth config settings should be read from a file.
                type: str
                choices:
                    - 'true'
                    - 'false'
            auth_file_path:
                description:
                    - The path of the config file containing auth settings.
                    - If the path is relative, base will the site's root directory.
                type: str
            config_version:
                description:
                    - The ConfigVersion of the Authentication / Authorization feature in use for the current app.
                    - The setting in this value can control the behavior of the control plane for Authentication / Authorization.
                type: str
    state:
        description:
            - State of the Web App.
            - Use C(present) to create or update a Web App and C(absent) to delete it.
        default: present
        type: str
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags
    - azure.azcollection.azure_identity_multiple

author:
    - Yunge Zhu (@yungezz)

'''

EXAMPLES = '''
- name: Create a windows web app with non-exist app service plan
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myWinWebapp
    plan:
      resource_group: myAppServicePlan_rg
      name: myAppServicePlan
      is_linux: false
      sku: S1

- name: Create a docker web app with some app settings, with docker image
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myDockerWebapp
    plan:
      resource_group: myAppServicePlan_rg
      name: myAppServicePlan
      is_linux: true
      sku: S1
      number_of_workers: 2
    app_settings:
      testkey: testvalue
      testkey2: testvalue2
    container_settings:
      name: ansible/ansible:ubuntu1404

- name: Create a docker web app with private acr registry
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myDockerWebapp
    plan: myAppServicePlan
    app_settings:
      testkey: testvalue
    container_settings:
      name: ansible/ubuntu1404
      registry_server_url: myregistry.io
      registry_server_user: user
      registry_server_password: "{{ password }}"

- name: Create a multi-container web app
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myMultiContainerWebapp
    plan: myAppServicePlan
    app_settings:
      testkey: testvalue
    container_settings:
      name: "COMPOSE|{{ lookup('file', 'docker-compose.yml') | b64encode }}"

- name: Create a linux web app with Node 6.6 framework
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myLinuxWebapp
    plan:
      resource_group: myAppServicePlan_rg
      name: myAppServicePlan
    app_settings:
      testkey: testvalue
    frameworks:
      - name: "node"
        version: "18"

- name: Create a windows web app with node, php
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myWinWebapp
    plan:
      resource_group: myAppServicePlan_rg
      name: myAppServicePlan
    app_settings:
      testkey: testvalue
    frameworks:
      - name: "node"
        version: 18
      - name: "php"
        version: 8.2

- name: Create a stage deployment slot for an existing web app
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myWebapp/slots/stage
    plan:
      resource_group: myAppServicePlan_rg
      name: myAppServicePlan
    app_settings:
      testkey:testvalue

- name: Create a linux web app with java framework
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myLinuxWebapp
    plan:
      resource_group: myAppServicePlan_rg
      name: myAppServicePlan
    app_settings:
      testkey: testvalue
    frameworks:
      - name: "java"
        version: "8"
        settings:
          java_container: "Tomcat"
          java_container_version: "8.5"

- name: Create a windows web app with site_auth_settings
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myWindowWebapp
    site_auth_settings:
      client_id: 'xxxxxxxx-xxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
      default_provider: 'MicrosoftAccount'
      runtime_version: '-2'
      token_refresh_extension_hours: 120
      unauthenticated_client_action: 'RedirectToLoginPage'
      client_secret: 'xxxxxxxx-xxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
      token_store_enabled: false
      enabled: true
      is_auth_from_file: false
    plan:
      resource_group: myResourceGroup
      name: myLinuxwebapp
      is_linux: false
      sku: S1

- name: Create a linux web app with python framework
  azure_rm_webapp:
    resource_group: myResourceGroup
    name: myLinuxWebapp
    plan:
      resource_group: myAppServicePlan_rg
      name: myAppServicePlan
    app_settings:
      testkey: testvalue
    frameworks:
      - name: "python"
        version: "3.10"
'''

RETURN = '''
id:
    description:
        - ID of current web app.
    returned: always
    type: str
    sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsoft.Web/sites/myWebApp"
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt

try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.core.polling import LROPoller
    from azure.core.exceptions import HttpResponseError
    from azure.mgmt.web.models import Site, AppServicePlan, SkuDescription, NameValuePair, SiteSourceControl, \
        StringDictionary, ManagedServiceIdentity, UserAssignedIdentity
except ImportError:
    # This is handled in azure_rm_common
    pass

container_settings_spec = dict(
    name=dict(type='str', required=True),
    registry_server_url=dict(type='str'),
    registry_server_user=dict(type='str'),
    registry_server_password=dict(type='str', no_log=True)
)

deployment_source_spec = dict(
    url=dict(type='str'),
    branch=dict(type='str')
)


framework_settings_spec = dict(
    java_container=dict(type='str', required=True),
    java_container_version=dict(type='str', required=True)
)


framework_spec = dict(
    name=dict(
        type='str',
        required=True,
        choices=['net_framework', 'java', 'php', 'node', 'python', 'dotnetcore', 'ruby']),
    version=dict(type='str', required=True),
    settings=dict(type='dict', options=framework_settings_spec)
)


site_auth_settings_spec = dict(
    kind=dict(type='str'),
    enabled=dict(type='bool'),
    runtime_version=dict(type='str'),
    unauthenticated_client_action=dict(type='str', choices=["RedirectToLoginPage", "AllowAnonymous"]),
    token_store_enabled=dict(type='bool'),
    allowed_external_redirect_urls=dict(type='list', elements='str'),
    default_provider=dict(type='str', choices=["Facebook", "Google", "MicrosoftAccount", "Twitter", "Github", "AzureActiveDirectory"]),
    token_refresh_extension_hours=dict(type='float', no_log=True),
    client_id=dict(type='str'),
    client_secret=dict(type='str', no_log=True),
    client_secret_setting_name=dict(type='str'),
    client_secret_certificate_thumbprint=dict(type='str', no_log=True),
    issuer=dict(type='str'),
    allowed_audiences=dict(type='list', elements='str'),
    additional_login_params=dict(type='str'),
    aad_claims_authorization=dict(type='str'),
    google_client_id=dict(type='str'),
    google_client_secret=dict(type='str', no_log=True),
    google_client_secret_setting_name=dict(type='str'),
    google_o_auth_scopes=dict(type='list', elements='str'),
    facebook_app_id=dict(type='str'),
    facebook_app_secret=dict(type='str', no_log=True),
    facebook_app_secret_setting_name=dict(type='str'),
    facebook_o_auth_scopes=dict(type='list', elements='str'),
    git_hub_client_id=dict(type='str'),
    git_hub_client_secret=dict(type='str', no_log=True),
    git_hub_client_secret_setting_name=dict(type='str'),
    git_hub_o_auth_scopes=dict(type='list', elements='str'),
    twitter_consumer_key=dict(type='str', no_log=True),
    twitter_consumer_secret=dict(type='str', no_log=True),
    twitter_consumer_secret_setting_name=dict(type='str'),
    microsoft_account_client_id=dict(type='str'),
    microsoft_account_client_secret=dict(type='str', no_log=True),
    microsoft_account_client_secret_setting_name=dict(type='str'),
    microsoft_account_o_auth_scopes=dict(type='list', elements='str'),
    is_auth_from_file=dict(type='str', choices=['true', 'false']),
    auth_file_path=dict(type='str'),
    config_version=dict(type='str')
)


def _normalize_sku(sku):
    if sku is None:
        return sku

    sku = sku.upper()
    if sku == 'FREE':
        return 'F1'
    elif sku == 'SHARED':
        return 'D1'
    return sku


def get_sku_name(tier):
    tier = tier.upper()
    if tier == 'F1' or tier == "FREE":
        return 'FREE'
    elif tier == 'D1' or tier == "SHARED":
        return 'SHARED'
    elif tier in ['B1', 'B2', 'B3', 'BASIC']:
        return 'BASIC'
    elif tier in ['S1', 'S2', 'S3']:
        return 'STANDARD'
    elif tier in ['P1', 'P2', 'P3']:
        return 'PREMIUM'
    elif tier in ['P1V2', 'P2V2', 'P3V2']:
        return 'PREMIUMV2'
    else:
        return None


def appserviceplan_to_dict(plan):
    return dict(
        id=plan.id,
        name=plan.name,
        kind=plan.kind,
        location=plan.location,
        reserved=plan.reserved,
        is_linux=plan.reserved,
        provisioning_state=plan.provisioning_state,
        tags=plan.tags if plan.tags else None
    )


def webapp_to_dict(webapp):
    return dict(
        id=webapp.id,
        name=webapp.name,
        location=webapp.location,
        client_cert_enabled=webapp.client_cert_enabled,
        enabled=webapp.enabled,
        reserved=webapp.reserved,
        client_affinity_enabled=webapp.client_affinity_enabled,
        server_farm_id=webapp.server_farm_id,
        host_names_disabled=webapp.host_names_disabled,
        https_only=webapp.https_only if hasattr(webapp, 'https_only') else None,
        skip_custom_domain_verification=webapp.skip_custom_domain_verification if hasattr(webapp, 'skip_custom_domain_verification') else None,
        ttl_in_seconds=webapp.ttl_in_seconds if hasattr(webapp, 'ttl_in_seconds') else None,
        state=webapp.state,
        identity=webapp.identity.as_dict() if webapp.identity else None,
        tags=webapp.tags if webapp.tags else None
    )


class Actions:
    CreateOrUpdate, UpdateAppSettings, UpdateAuthSettings, Delete = range(4)


class AzureRMWebApps(AzureRMModuleBaseExt):
    """Configuration class for an Azure RM Web App resource"""

    def __init__(self):
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
                required=True
            ),
            name=dict(
                type='str',
                required=True
            ),
            location=dict(
                type='str'
            ),
            plan=dict(
                type='raw'
            ),
            frameworks=dict(
                type='list',
                elements='dict',
                options=framework_spec
            ),
            container_settings=dict(
                type='dict',
                options=container_settings_spec
            ),
            scm_type=dict(
                type='str',
            ),
            always_on=dict(
                type='bool',
            ),
            http20_enabled=dict(
                type='bool',
            ),
            min_tls_version=dict(
                type='str',
                choices=['1.0', '1.1', '1.2'],
            ),
            ftps_state=dict(
                type='str',
                choices=['AllAllowed', 'FtpsOnly', 'Disabled'],
            ),
            deployment_source=dict(
                type='dict',
                options=deployment_source_spec
            ),
            startup_file=dict(
                type='str'
            ),
            client_affinity_enabled=dict(
                type='bool',
                default=True
            ),
            https_only=dict(
                type='bool'
            ),
            app_settings=dict(
                type='dict'
            ),
            purge_app_settings=dict(
                type='bool',
                default=False
            ),
            app_state=dict(
                type='str',
                choices=['started', 'stopped', 'restarted'],
                default='started'
            ),
            site_auth_settings=dict(
                type='dict',
                options=site_auth_settings_spec
            ),
            identity=dict(
                type='dict',
                options=self.managed_identity_multiple_spec
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            )
        )

        mutually_exclusive = [['container_settings', 'frameworks']]

        self.resource_group = None
        self.name = None
        self.location = None

        # update in create_or_update as parameters
        self.client_affinity_enabled = True
        self.https_only = None

        self.tags = None

        # Managed Identity
        self.identity = None

        # site config, e.g app settings, ssl
        self.site_config = dict()
        self.app_settings = dict()
        self.app_settings_strDic = None
        self.site_auth_settings = None

        # app service plan
        self.plan = None

        # siteSourceControl
        self.deployment_source = dict()

        # site, used at level creation, or update. e.g windows/linux, client_affinity etc first level args
        self.site = None

        # property for internal usage, not used for sdk
        self.container_settings = None

        self.purge_app_settings = False
        self.app_state = 'started'

        self.results = dict(
            changed=False,
            id=None,
        )
        self.state = None
        self.to_do = []

        self.frameworks = None

        # set site_config value from kwargs
        self.site_config_updatable_properties = ["net_framework_version",
                                                 "java_version",
                                                 "php_version",
                                                 "python_version",
                                                 "scm_type",
                                                 "always_on",
                                                 "http20_enabled",
                                                 "min_tls_version",
                                                 "ftps_state",
                                                 "app_command_line"]

        # updatable_properties
        self.updatable_properties = ["client_affinity_enabled",
                                     "https_only"]

        self.supported_linux_frameworks = ['ruby', 'php', 'python', 'dotnetcore', 'node', 'java']
        self.supported_windows_frameworks = ['net_framework', 'php', 'python', 'node', 'java', 'dotnetcore']

        self._managed_identity = None

        self.startup_file = None

        super(AzureRMWebApps, self).__init__(derived_arg_spec=self.module_arg_spec,
                                             mutually_exclusive=mutually_exclusive,
                                             supports_check_mode=True,
                                             supports_tags=True)

    @property
    def managed_identity(self):
        if not self._managed_identity:
            self._managed_identity = {"identity": ManagedServiceIdentity,
                                      "user_assigned": UserAssignedIdentity
                                      }
        return self._managed_identity

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
            elif kwargs[key] is not None:
                if key in ["scm_type", "always_on", "http20_enabled", "min_tls_version", "ftps_state"]:
                    self.site_config[key] = kwargs[key]

        old_response = None
        response = None
        to_be_updated = False

        # set location
        resource_group = self.get_resource_group(self.resource_group)
        if not self.location:
            self.location = resource_group.location

        # get existing web app
        old_response = self.get_webapp()

        if old_response:
            self.results['id'] = old_response['id']
            self.results['site_auth_settings'] = self.get_auth_settings()

        if self.state == 'present':
            if not self.plan and not old_response:
                self.fail("Please specify plan for newly created web app.")

            if not self.plan:
                self.plan = old_response['server_farm_id']

            self.plan = self.parse_resource_to_dict(self.plan)

            # get app service plan
            is_linux = False
            old_plan = self.get_app_service_plan()
            if old_plan:
                is_linux = old_plan['reserved']
            else:
                is_linux = self.plan['is_linux'] if 'is_linux' in self.plan else False

            if self.frameworks:
                # java is mutually exclusive with other frameworks
                if len(self.frameworks) > 1 and any(f['name'] == 'java' for f in self.frameworks):
                    self.fail('Java is mutually exclusive with other frameworks.')

                if is_linux:
                    if len(self.frameworks) != 1:
                        self.fail('Can specify one framework only for Linux web app.')

                    if self.frameworks[0]['name'] not in self.supported_linux_frameworks:
                        self.fail('Unsupported framework {0} for Linux web app.'.format(self.frameworks[0]['name']))

                    self.site_config['linux_fx_version'] = (self.frameworks[0]['name'] + '|' + self.frameworks[0]['version']).upper()

                    if self.frameworks[0]['name'] == 'java':
                        if self.frameworks[0]['version'] not in ['8', '11', '17', '21']:
                            self.fail("Linux web app only supports java 8, 11, 17 and 21.")
                        if self.frameworks[0]['settings'] and self.frameworks[0]['settings']['java_container'].lower() != 'tomcat':
                            self.fail("Linux web app only supports tomcat container.")

                        if self.frameworks[0]['settings'] and self.frameworks[0]['settings']['java_container'].lower() == 'tomcat':
                            self.site_config['linux_fx_version'] = 'TOMCAT|' + self.frameworks[0]['settings']['java_container_version'] + '-jre8'
                        else:
                            self.site_config['linux_fx_version'] = 'JAVA|{0}-jre{0}'.format(self.frameworks[0]['version'])
                else:
                    for fx in self.frameworks:
                        if fx.get('name') not in self.supported_windows_frameworks:
                            self.fail('Unsupported framework {0} for Windows web app.'.format(fx.get('name')))
                        else:
                            self.site_config[fx.get('name') + '_version'] = fx.get('version')

                        if 'settings' in fx and fx['settings'] is not None:
                            for key, value in fx['settings'].items():
                                self.site_config[key] = value

            if not self.app_settings:
                self.app_settings = dict()

            if self.container_settings:
                linux_fx_version = 'DOCKER|'

                if self.container_settings.get('registry_server_url'):
                    self.app_settings['DOCKER_REGISTRY_SERVER_URL'] = 'https://' + self.container_settings['registry_server_url']

                    linux_fx_version += self.container_settings['registry_server_url'] + '/'

                linux_fx_version += self.container_settings['name']

                # Use given name as is if it starts with allowed values of multi-container application
                if self.container_settings['name'].startswith('COMPOSE|') or self.container_settings['name'].startswith('KUBE|'):
                    linux_fx_version = self.container_settings['name']

                self.site_config['linux_fx_version'] = linux_fx_version

                if self.container_settings.get('registry_server_user'):
                    self.app_settings['DOCKER_REGISTRY_SERVER_USERNAME'] = self.container_settings['registry_server_user']

                if self.container_settings.get('registry_server_password'):
                    self.app_settings['DOCKER_REGISTRY_SERVER_PASSWORD'] = self.container_settings['registry_server_password']

            # init site
            self.site = Site(location=self.location, site_config=self.site_config)

            if self.https_only is not None:
                self.site.https_only = self.https_only

            self.site.client_affinity_enabled = self.client_affinity_enabled

            # check if the web app already present in the resource group
            if not old_response:
                self.log("Web App instance doesn't exist")

                to_be_updated = True
                self.to_do.append(Actions.CreateOrUpdate)
                if self.site_auth_settings is not None:
                    self.to_do.append(Actions.UpdateAuthSettings)
                self.site.tags = self.tags

                if self.identity:
                    update_identity, self.site.identity = self.update_managed_identity(new_identity=self.identity)

                # service plan is required for creation
                if not self.plan:
                    self.fail("Please specify app service plan in plan parameter.")

                if not old_plan:
                    # no existing service plan, create one
                    if (not self.plan.get('name') or not self.plan.get('sku')):
                        self.fail('Please specify name, is_linux, sku in plan')

                    if 'location' not in self.plan:
                        plan_resource_group = self.get_resource_group(self.plan['resource_group'])
                        self.plan['location'] = plan_resource_group.location

                    old_plan = self.create_app_service_plan()

                self.site.server_farm_id = old_plan['id']

                # if linux, setup startup_file
                if old_plan['is_linux']:
                    if hasattr(self, 'startup_file'):
                        self.site_config['app_command_line'] = self.startup_file

                # set app setting
                if self.app_settings:
                    app_settings = []
                    for key in self.app_settings.keys():
                        app_settings.append(NameValuePair(name=key, value=self.app_settings[key]))

                    self.site_config['app_settings'] = app_settings
            else:
                # existing web app, do update
                self.log("Web App instance already exists")

                self.log('Result: {0}'.format(old_response))

                if self.identity:
                    update_identity, self.site.identity = self.update_managed_identity(new_identity=self.identity,
                                                                                       curr_identity=old_response.get('identity', None))

                    if update_identity:
                        to_be_updated = True
                        self.to_do.append(Actions.CreateOrUpdate)

                update_tags, self.site.tags = self.update_tags(old_response.get('tags', None))

                if update_tags:
                    to_be_updated = True
                    self.to_do.append(Actions.CreateOrUpdate)

                # check if root level property changed
                if self.is_updatable_property_changed(old_response):
                    to_be_updated = True
                    self.to_do.append(Actions.CreateOrUpdate)

                # check if site_config changed
                old_config = self.get_webapp_configuration()

                if self.is_startup_file_changed(old_config):
                    self.site_config['app_command_line'] = self.startup_file

                if self.is_site_config_changed(old_config):
                    to_be_updated = True
                    self.to_do.append(Actions.CreateOrUpdate)

                # check if linux_fx_version changed
                if old_config.linux_fx_version != self.site_config.get('linux_fx_version', ''):
                    to_be_updated = True
                    self.to_do.append(Actions.CreateOrUpdate)

                self.app_settings_strDic = self.list_app_settings()

                # purge existing app_settings:
                if self.purge_app_settings:
                    to_be_updated = True
                    self.app_settings_strDic = dict()
                    self.to_do.append(Actions.UpdateAppSettings)

                # check if app settings changed
                if self.purge_app_settings or self.is_app_settings_changed():
                    to_be_updated = True
                    self.to_do.append(Actions.UpdateAppSettings)

                    if self.app_settings:
                        for key in self.app_settings.keys():
                            self.app_settings_strDic[key] = self.app_settings[key]

                if self.site_auth_settings is not None:
                    result = dict(compare=[])
                    if not self.default_compare({}, self.site_auth_settings, self.results['site_auth_settings'], '', dict(compare=[])):
                        to_be_updated = True
                        self.to_do.append(Actions.UpdateAuthSettings)

        elif self.state == 'absent':
            if old_response:
                self.log("Delete Web App instance")
                self.results['changed'] = True

                if self.check_mode:
                    return self.results

                self.delete_webapp()

                self.log('Web App instance deleted')

            else:
                self.log("Web app {0} not exists.".format(self.name))

        if to_be_updated:
            self.log('Need to Create/Update web app')
            self.results['changed'] = True

            if self.check_mode:
                return self.results

            if Actions.CreateOrUpdate in self.to_do:
                response = self.create_update_webapp()

                self.results['id'] = response['id']

            if Actions.UpdateAppSettings in self.to_do:
                update_response = self.update_app_settings()
                self.results['id'] = update_response.id

            if Actions.UpdateAuthSettings in self.to_do:
                auth_settings = self.update_auth_settings(self.site_auth_settings)
                self.results['site_auth_settings'] = auth_settings
            else:
                self.results['site_auth_settings'] = self.get_auth_settings()

        webapp = None
        if old_response:
            webapp = old_response
        if response:
            webapp = response

        if webapp:
            if (webapp['state'] != 'Stopped' and self.app_state == 'stopped') or \
               (webapp['state'] != 'Running' and self.app_state == 'started') or \
               self.app_state == 'restarted':

                self.results['changed'] = True
                if self.check_mode:
                    return self.results

                self.set_webapp_state(self.app_state)

        return self.results

    # compare existing web app with input, determine weather it's update operation
    def is_updatable_property_changed(self, existing_webapp):
        for property_name in self.updatable_properties:
            if hasattr(self, property_name) and getattr(self, property_name) is not None and \
                    getattr(self, property_name) != existing_webapp.get(property_name, None):
                return True

        return False

    # compare xxx_version
    def is_site_config_changed(self, existing_config):
        for updatable_property in self.site_config_updatable_properties:
            if updatable_property in self.site_config:
                if not getattr(existing_config, updatable_property) or \
                        str(getattr(existing_config, updatable_property)).upper() != str(self.site_config.get(updatable_property)).upper():
                    return True

        return False

    # check if startup file changed
    def is_startup_file_changed(self, existing_config):
        self_name = "startup_file"
        property_name = "app_command_line"
        if hasattr(self, self_name) and getattr(self, self_name) is not None:
            if not getattr(existing_config, property_name) or \
                    str(getattr(existing_config, property_name)).upper() != str(self.startup_file).upper():
                return True

        return False

    # comparing existing app setting with input, determine whether it's changed
    def is_app_settings_changed(self):
        if self.app_settings:
            if self.app_settings_strDic:
                for key in self.app_settings.keys():
                    if self.app_settings[key] != self.app_settings_strDic.get(key, None):
                        return True
            else:
                return True
        return False

    # comparing deployment source with input, determine wheather it's changed
    def is_deployment_source_changed(self, existing_webapp):
        if self.deployment_source:
            if self.deployment_source.get('url') \
                    and self.deployment_source['url'] != existing_webapp.get('site_source_control')['url']:
                return True

            if self.deployment_source.get('branch') \
                    and self.deployment_source['branch'] != existing_webapp.get('site_source_control')['branch']:
                return True

        return False

    def create_update_webapp(self):
        '''
        Creates or updates Web App with the specified configuration.

        :return: deserialized Web App instance state dictionary
        '''
        self.log(
            "Creating / Updating the Web App instance {0}".format(self.name))

        try:
            response = self.web_client.web_apps.begin_create_or_update(resource_group_name=self.resource_group,
                                                                       name=self.name,
                                                                       site_envelope=self.site)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)

        except Exception as exc:
            self.log('Error attempting to create the Web App instance.')
            self.fail("Error creating the Web App instance: {0}".format(str(exc)))
        return webapp_to_dict(response)

    def delete_webapp(self):
        '''
        Deletes specified Web App instance in the specified subscription and resource group.

        :return: True
        '''
        self.log("Deleting the Web App instance {0}".format(self.name))
        try:
            self.web_client.web_apps.delete(resource_group_name=self.resource_group, name=self.name)
        except Exception as e:
            self.log('Error attempting to delete the Web App instance.')
            self.fail("Error deleting the Web App instance: {0}".format(str(e)))

        return True

    def get_webapp(self):
        '''
        Gets the properties of the specified Web App.

        :return: deserialized Web App instance state dictionary
        '''
        self.log(
            "Checking if the Web App instance {0} is present".format(self.name))

        response = None

        try:
            response = self.web_client.web_apps.get(resource_group_name=self.resource_group, name=self.name)

            # Newer SDK versions (0.40.0+) seem to return None if it doesn't exist instead of raising error
            if response is not None:
                self.log("Response : {0}".format(response))
                self.log("Web App instance : {0} found".format(response.name))
                return webapp_to_dict(response)

        except ResourceNotFoundError:
            pass

        self.log("Didn't find web app {0} in resource group {1}".format(self.name, self.resource_group))

        return False

    def get_app_service_plan(self):
        '''
        Gets app service plan
        :return: deserialized app service plan dictionary
        '''
        self.log("Get App Service Plan {0}".format(self.plan['name']))

        try:
            response = self.web_client.app_service_plans.get(
                resource_group_name=self.plan['resource_group'],
                name=self.plan['name'])

            # Newer SDK versions (0.40.0+) seem to return None if it doesn't exist instead of raising error
            if response is not None:
                self.log("Response : {0}".format(response))
                self.log("App Service Plan : {0} found".format(response.name))

                return appserviceplan_to_dict(response)
        except ResourceNotFoundError:
            pass

        self.log("Didn't find app service plan {0} in resource group {1}".format(
            self.plan['name'], self.plan['resource_group']))

        return False

    def create_app_service_plan(self):
        '''
        Creates app service plan
        :return: deserialized app service plan dictionary
        '''
        self.log("Create App Service Plan {0}".format(self.plan['name']))

        try:
            # normalize sku
            sku = _normalize_sku(self.plan['sku'])

            sku_def = SkuDescription(tier=get_sku_name(
                sku), name=sku, capacity=(self.plan.get('number_of_workers', None)))
            plan_def = AppServicePlan(
                location=self.plan['location'], app_service_plan_name=self.plan['name'], sku=sku_def, reserved=(self.plan.get('is_linux', None)))

            poller = self.web_client.app_service_plans.begin_create_or_update(
                resource_group_name=self.plan['resource_group'], name=self.plan['name'], app_service_plan=plan_def)

            if isinstance(poller, LROPoller):
                response = self.get_poller_result(poller)

            self.log("Response : {0}".format(response))

            return appserviceplan_to_dict(response)
        except Exception as ex:
            self.fail("Failed to create app service plan {0} in resource group {1}: {2}".format(
                self.plan['name'], self.plan['resource_group'], str(ex)))

    def list_app_settings(self):
        '''
        List application settings
        :return: deserialized list response
        '''
        self.log("List application setting")

        try:
            response = self.web_client.web_apps.list_application_settings(resource_group_name=self.resource_group, name=self.name)
            self.log("Response : {0}".format(response))

            return response.properties
        except Exception as ex:
            self.fail("Failed to list application settings for web app {0} in resource group {1}: {2}".format(
                self.name, self.resource_group, str(ex)))

    def update_auth_settings(self, body):
        '''
        Update web app auth settings
        :return: deserialized updating response
        '''
        self.log("Update web app auth settings")

        try:
            response = self.web_client.web_apps.update_auth_settings(self.resource_group, self.name, body)
            self.log("Response : {0}".format(response))
            return response.as_dict()
        except HttpResponseError as ex:
            self.fail("Failed to update web app auth settings for web app {0} in resource group {1}: {2}".format(
                self.name, self.resource_group, str(ex)))

    def get_auth_settings(self):
        '''
        Get the web app auth settings
        :return: deserialized updating response
        '''
        self.log("Get the web app auth settings")

        try:
            response = self.web_client.web_apps.get_auth_settings(self.resource_group, self.name)
            self.log("Response : {0}".format(response))
            return response.as_dict()
        except HttpResponseError as ex:
            self.log("Failed to Get web app auth settings for web app {0} in resource group {1}: {2}".format(
                     self.name, self.resource_group, str(ex)))

    def update_app_settings(self):
        '''
        Update application settings
        :return: deserialized updating response
        '''
        self.log("Update application setting")

        try:
            settings = StringDictionary(
                properties=self.app_settings_strDic
            )
            response = self.web_client.web_apps.update_application_settings(
                resource_group_name=self.resource_group, name=self.name, app_settings=settings)
            self.log("Response : {0}".format(response))
            return response
        except HttpResponseError as ex:
            self.fail("Failed to update application settings for web app {0} in resource group {1}: {2}".format(
                self.name, self.resource_group, str(ex)))

    def create_or_update_source_control(self):
        '''
        Update site source control
        :return: deserialized updating response
        '''
        self.log("Update site source control")

        if self.deployment_source is None:
            return False

        self.deployment_source['is_manual_integration'] = False
        self.deployment_source['is_mercurial'] = False

        try:
            site_source_control = SiteSourceControl(
                repo_url=self.deployment_source.get('url'),
                branch=self.deployment_source.get('branch')
            )
            response = self.web_client.web_apps.begin_create_or_update_source_control(
                resource_group_name=self.resource_group, name=self.name, site_source_control=site_source_control)
            self.log("Response : {0}".format(response))

            return response.as_dict()
        except Exception:
            self.fail("Failed to update site source control for web app {0} in resource group {1}".format(
                self.name, self.resource_group))

    def get_webapp_configuration(self):
        '''
        Get  web app configuration
        :return: deserialized  web app configuration response
        '''
        self.log("Get web app configuration")

        try:
            response = self.web_client.web_apps.get_configuration(
                resource_group_name=self.resource_group, name=self.name)
            self.log("Response : {0}".format(response))

            return response
        except ResourceNotFoundError as ex:
            self.log("Failed to get configuration for web app {0} in resource group {1}: {2}".format(
                self.name, self.resource_group, str(ex)))

            return False

    def set_webapp_state(self, appstate):
        '''
        Start/stop/restart web app
        :return: deserialized updating response
        '''
        try:
            if appstate == 'started':
                response = self.web_client.web_apps.start(resource_group_name=self.resource_group, name=self.name)
            elif appstate == 'stopped':
                response = self.web_client.web_apps.stop(resource_group_name=self.resource_group, name=self.name)
            elif appstate == 'restarted':
                response = self.web_client.web_apps.restart(resource_group_name=self.resource_group, name=self.name)
            else:
                self.fail("Invalid web app state {0}".format(appstate))

            self.log("Response : {0}".format(response))

            return response
        except Exception as ex:
            request_id = ex.request_id if ex.request_id else ''
            self.log("Failed to {0} web app {1} in resource group {2}, request_id {3} - {4}".format(
                appstate, self.name, self.resource_group, request_id, str(ex)))


def main():
    """Main execution"""
    AzureRMWebApps()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_vpc_route_table_info
version_added: 1.0.0
short_description: Gather information about ec2 VPC route tables in AWS
description:
    - Gather information about EC2 VPC route tables in AWS.
author:
- "Rob White (@wimnat)"
- "Mark Chappell (@tremble)"
options:
  filters:
    description:
      - A dict of filters to apply. Each dict item consists of a filter key and a filter value.
        See U(https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeRouteTables.html) for possible filters.
    type: dict
    default: {}
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Gather information about all VPC route tables
  amazon.aws.ec2_vpc_route_table_info:

- name: Gather information about a particular VPC route table using route table ID
  amazon.aws.ec2_vpc_route_table_info:
    filters:
      route-table-id: rtb-00112233

- name: Gather information about any VPC route table with a tag key Name and value Example
  amazon.aws.ec2_vpc_route_table_info:
    filters:
      "tag:Name": Example

- name: Gather information about any VPC route table within VPC with ID vpc-abcdef00
  amazon.aws.ec2_vpc_route_table_info:
    filters:
      vpc-id: vpc-abcdef00
"""

RETURN = r"""
route_tables:
  description:
    - A list of dictionarys describing route tables.
    - See also U(https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_route_tables).
  returned: always
  type: complex
  contains:
    associations:
      description: List of associations between the route table and one or more subnets or a gateway.
      returned: always
      type: complex
      contains:
        association_state:
          description: The state of the association.
          returned: always
          type: complex
          contains:
            state:
              description: The state of the association.
              returned: always
              type: str
              sample: associated
            state_message:
              description: Additional information about the state of the association.
              returned: when available
              type: str
              sample: 'Creating association'
        gateway_id:
          description: ID of the internet gateway or virtual private gateway.
          returned: when route table is a gateway route table
          type: str
          sample: igw-03312309
        main:
          description: Whether this is the main route table.
          returned: always
          type: bool
          sample: false
        route_table_association_id:
          description: ID of association between route table and subnet.
          returned: always
          type: str
          sample: rtbassoc-ab47cfc3
        route_table_id:
          description: ID of the route table.
          returned: always
          type: str
          sample: rtb-bf779ed7
        subnet_id:
          description: ID of the subnet.
          returned: when route table is a subnet route table
          type: str
          sample: subnet-82055af9
    id:
      description: ID of the route table (same as RV(route_tables.route_table_id) for backwards compatibility).
      returned: always
      type: str
      sample: rtb-bf779ed7
    owner_id:
      description: ID of the account which owns the route table.
      returned: always
      type: str
      sample: '012345678912'
    propagating_vgws:
      description: List of Virtual Private Gateways propagating routes.
      returned: always
      type: list
      sample: []
    route_table_id:
      description: ID of the route table.
      returned: always
      type: str
      sample: rtb-bf779ed7
    routes:
      description: List of routes in the route table.
      returned: always
      type: complex
      contains:
        destination_cidr_block:
          description: CIDR block of destination.
          returned: always
          type: str
          sample: 10.228.228.0/22
        gateway_id:
          description: ID of the gateway.
          returned: when gateway is local or internet gateway
          type: str
          sample: local
        instance_id:
          description:
            - ID of a NAT instance.
            - Empty unless the route is via an EC2 instance.
          returned: always
          type: str
          sample: i-abcd123456789
        instance_owner_id:
          description:
            - AWS account owning the NAT instance.
            - Empty unless the route is via an EC2 instance.
          returned: always
          type: str
          sample: 123456789012
        network_interface_id:
          description:
            - The ID of the network interface.
            - Empty unless the route is via an EC2 instance.
          returned: always
          type: str
          sample: 123456789012
        nat_gateway_id:
          description: ID of the NAT gateway.
          returned: when the route is via a NAT gateway.
          type: str
          sample: local
        transit_gateway_id:
          description: ID of the Transit gateway.
          returned: when the route is via a Transit gateway.
          type: str
          sample: tgw-123456789012
          version_added: 9.0.0
        origin:
          description: mechanism through which the route is in the table.
          returned: always
          type: str
          sample: CreateRouteTable
        state:
          description: state of the route.
          returned: always
          type: str
          sample: active
    tags:
      description: Tags applied to the route table.
      returned: always
      type: dict
      sample:
        Name: Public route table
        Public: 'true'
    vpc_id:
      description: ID for the VPC in which the route lives.
      returned: always
      type: str
      sample: vpc-6e2d2407
"""

from typing import Any
from typing import Dict

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_route_tables
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list


def normalize_route(route: Dict[str, Any]) -> Dict[str, Any]:
    # Historically these were all there, but set to null when empty'
    for legacy_key in ["DestinationCidrBlock", "GatewayId", "InstanceId", "Origin", "State", "NetworkInterfaceId"]:
        if legacy_key not in route:
            route[legacy_key] = None
    route["InterfaceId"] = route["NetworkInterfaceId"]
    return route


def normalize_association(assoc: Dict[str, Any]) -> Dict[str, Any]:
    # Name change between boto v2 and boto v3, return both
    assoc["Id"] = assoc["RouteTableAssociationId"]
    return assoc


def normalize_route_table(table: Dict[str, Any]) -> Dict[str, Any]:
    table["tags"] = boto3_tag_list_to_ansible_dict(table["Tags"])
    table["Associations"] = [normalize_association(assoc) for assoc in table["Associations"]]
    table["Routes"] = [normalize_route(route) for route in table["Routes"]]
    table["Id"] = table["RouteTableId"]
    del table["Tags"]
    return camel_dict_to_snake_dict(table, ignore_list=["tags"])


def list_ec2_vpc_route_tables(connection, module: AnsibleAWSModule) -> None:
    filters = ansible_dict_to_boto3_filter_list(module.params.get("filters"))

    try:
        route_tables = describe_route_tables(connection, Filters=filters)
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)

    route_tables = [normalize_route_table(table) for table in route_tables]
    module.exit_json(changed=False, route_tables=route_tables)


def main() -> None:
    argument_spec = dict(
        filters=dict(default={}, type="dict"),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)

    connection = module.client("ec2")

    list_ec2_vpc_route_tables(connection, module)


if __name__ == "__main__":
    main()
